# pylint: skip-file
import typing

import pytest_asyncio
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken

from cbra.ext.oauth2 import ConfigFileClientRepository
from cbra.ext.oauth2.types import ClientAssertionType
from cbra.ext.oauth2.types import IClient


class AssertedClientMixin:
    assertion_type: ClientAssertionType = ClientAssertionType.jwt_bearer
    client_id: str
    client_key: str
    token_url: str

    async def get_client_keychain(self, key: typing.Any) -> Keychain:
        raise NotImplementedError

    @pytest_asyncio.fixture # type: ignore
    async def client_keychain(self) -> Keychain:
        return await self.get_client_keychain(self.client_key)

    @pytest_asyncio.fixture # type: ignore
    async def current_client(self) -> IClient:
        return await ConfigFileClientRepository().get(self.client_id)

    async def get_client_assertion(
        self,
        keychain: Keychain,
        **claims: typing.Any
    ) -> str:
        codec = PayloadCodec(signer=keychain)
        claims = {
            'iss': self.client_id,
            'aud': self.token_url,
            'ttl': 30,
            **claims
        }
        return await codec.encode(
            JSONWebToken.strict(**claims),
            signers=['sig']
        )

    @pytest_asyncio.fixture # type: ignore
    async def assertion(
        self,
        client_keychain: Keychain
    ) -> str:
        return await self.get_client_assertion(client_keychain)



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from ckms.core import Keychain
from ckms.core.models import JSONWebSignature

import cbra
from cbra.parsers import IParser
from cbra.parsers import JWTParser
from cbra.ext.jwks import JWKSEndpoint
from cbra.ext.rek import RequestEncryptionKeyEndpoint


KEYCHAIN: dict[str, Any] = {
    'request-encryption-key': {
        'provider': 'local',
        'kty': 'RSA',
        'algorithm': 'RSA-OAEP-256',
        'key': {'path': 'pki/rsa-enc.key'},
        'tags': ['request-encryption-key']
    },
    'reks-signer-1': {
        'provider': 'local',
        'kty': 'OKP',
        'curve': 'Ed25519',
        'key': {'path': 'pki/ed25519.key'},
        'tags': ['reks-signer', 'unimatrixone.io/public-keys']
    },
    'reks-signer-2': {
        'provider': 'local',
        'kty': 'EC',
        'use': 'sig',
        'key': {'path': 'pki/p384.key'},
        'tags': ['reks-signer', 'unimatrixone.io/public-keys']
    },
}

keychain: Keychain = Keychain()
keychain.configure(KEYCHAIN)


class EncryptedBodyEndpoint(cbra.Endpoint):
    method: str = 'POST'
    parsers: list[type[IParser]] = [
        JWTParser
    ]

    async def handle( # type: ignore
        self,
        jws: JSONWebSignature
    ) -> dict[str, Any]:
        jwt = jws.claims({"jwt"})
        return jwt.dict()


def get_asgi_application() -> cbra.Application:
    app: cbra.Application = cbra.Application(keychain=keychain)
    app.add(RequestEncryptionKeyEndpoint)
    app.add(EncryptedBodyEndpoint, base_path='/')
    app.add(JWKSEndpoint, base_path='/')
    return app


app = get_asgi_application()


if __name__ == '__main__':
    cbra.run('__main__:app', reload=True)


# pylint: skip-file
# type: ignore
import types

import pytest

from cbra.ext.oauth2.types import IClientRepository


@pytest.fixture
def settings():
    settings = types.ModuleType('settings')
    setattr(settings, 'OAUTH_CLIENTS', None)
    return settings


def test_configure_from_string(settings: types.ModuleType):
    settings.OAUTH_CLIENTS = 'cbra.ext.oauth2.SettingsClientRepository'
    ClientRepository = IClientRepository.fromsettings(settings)


def test_configure_from_string_with_attrs(settings: types.ModuleType):
    settings.OAUTH_CLIENTS = {
        'class': 'cbra.ext.oauth2.SettingsClientRepository',
        'foo': 'bar'
    }
    ClientRepository = IClientRepository.fromsettings(settings)
    assert hasattr(ClientRepository, 'foo')
    assert ClientRepository.foo == 'bar'


"""Declares :class:`EmptyResponse`."""
from typing import Any

from starlette.responses import Response


class EmptyResponse(Response):
    __module__: str = 'cbra.responses'

    def __init__(self, **kwargs: Any):
        kwargs['status_code'] = 204
        super().__init__(**kwargs)


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
import secrets
from typing import Any

from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.utils import current_timestamp

from cbra.params import ServerKeychain
from .params import ServerCodec
from .params import CurrentServerMetadata
from .oidcclaimhandler import OIDCClaimHandler
from .types import AuthorizationRequest
from .types import AuthorizationRequestClaims
from .types import BaseGrant
from .types import AuthorizationCodeGrant
from .types import IClient
from .types import IOIDCTokenBuilder
from .types import ISubject
from .types import ServerMetadata
from .utils import openid_hash


class OIDCTokenBuilder(IOIDCTokenBuilder):
    __module__: str = 'cbra.ext.oauth2'
    claims_model: type[AuthorizationRequestClaims] = AuthorizationRequestClaims
    codec: PayloadCodec
    handlers: list[type[OIDCClaimHandler]] = []
    keychain: Keychain
    metadata:  ServerMetadata
    now: int

    def __init__(
        self,
        codec: PayloadCodec = ServerCodec,
        keychain: Keychain = ServerKeychain,
        metadata: ServerMetadata = CurrentServerMetadata
    ):
        self.codec = codec
        self.keychain = keychain
        self.metadata = metadata
        self.now = current_timestamp()

    def get_handlers(
        self,
        client: IClient,
        subject: ISubject,
        initial: dict[str, Any] | None = None,
        scope: set[str] | None = None
    ) -> list[OIDCClaimHandler]:
        """Return a list of handler instances to construct the claims
        set.
        """
        scope = scope or set()
        handlers: list[OIDCClaimHandler] = []
        for cls in self.handlers:
            handler = cls(
                client,
                subject,
                self.now,
                copy.deepcopy(initial or {})
            )
            if not handler.is_enabled(scope):
                continue
            handlers.append(handler)
        return handlers

    def get_initial_claims(
        self,
        signing_key: str,
        client: IClient,
        subject: ISubject,
        grant: BaseGrant,
        request: AuthorizationRequest | None = None,
        access_token: str | None = None,
        code: str | None = None
    ) -> dict[str, Any]:
        """Return a dictionary containing the initial claims to add to the
        ID Token.
        """
        key = self.keychain.get(signing_key)
        claims: dict[str, Any] = {
            **client.get_id_token_claims(self.now),
            'iat': self.now,
            'iss': self.metadata.issuer,
            'jti': secrets.token_urlsafe(24),
            'nbf': self.now,
            'sub': client.get_subject_id(subject)
        }
        if isinstance(grant, AuthorizationCodeGrant):
            assert isinstance(request, AuthorizationRequest) # nosec
            assert isinstance(request.auth_time, int)
            claims.update({
                'auth_time': request.auth_time,
                'c_hash': openid_hash(
                    alg=key.algorithm,
                    value=grant.code,
                    crv=key.curve
                )
            })
        if access_token is not None:
            claims['at_hash'] = openid_hash(
                alg=key.algorithm,
                value=access_token,
                crv=key.curve
            )
        if request is not None:
            claims["nonce"] = request.nonce
        return claims

    def requests(
        self,
        scope: set[str]
    ) -> set[str]:
        claims: set[str] = set()
        for handler in self.handlers:
            claims.update(handler.requests(scope))
        return claims

    async def build(
        self,
        *,
        signing_key: str,
        client: IClient,
        subject: ISubject,
        grant: BaseGrant,
        scope: set[str],
        request: AuthorizationRequest | None = None,
        access_token: str | None = None
    ) -> str | None:
        if not self._is_openid(grant, request):
            return None
        claims: dict[str, Any] = self.get_initial_claims(
            signing_key=signing_key,
            client=client,
            subject=subject,
            request=request,
            grant=grant,
            access_token=access_token
        )
        return await self.codec.encode(
            payload=await self.run_handlers(
                client=client,
                subject=subject,
                initial=claims,
                scope=scope,
                grant=grant,
                request=request
            ),
            signers=[signing_key],
            encrypters=client.get_encryption_keys() # type: ignore
        )

    async def run_handlers(
        self,
        client: IClient,
        subject: ISubject,
        initial: dict[str, Any] | None = None,
        scope: set[str] | None = None,
        grant: BaseGrant | None = None,
        request: AuthorizationRequest | None = None
    ) -> dict[str, Any]:
        """Run handlers to update the initial claims with their respective
        additions.
        """
        initial = initial or {}
        scope = scope or set()
        claims: dict[str, Any] = {}
        for handler in self.get_handlers(client, subject, initial, scope):
            claims.update(await handler.claims(grant, request))
        return {**initial, **claims}

    def _is_openid(
        self,
        grant: BaseGrant,
        request: AuthorizationRequest | None
    ) -> bool:
        return bool(grant.is_openid() or (request and request.is_openid()))


# pylint: skip-file
import typing

import fastapi
import pytest
import pytest_asyncio
from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebKey
from httpx import AsyncClient

import cbra
import cbra.auth
from cbra.ext.oauth2 import ref
from cbra.ext.oauth2 import AuthorizationEndpoint
from cbra.ext.oauth2 import AuthorizationServer
from cbra.ext.oauth2 import ConfigFileClientRepository
from cbra.ext.oauth2 import MemoryStorage
from cbra.ext.oauth2 import MemorySubjectRepository
from cbra.ext.oauth2 import RFC9068Principal
from cbra.ext.oauth2.types import IPrincipal
from cbra.ext.oauth2.types import ISubject
from cbra.ext.oauth2.types import ISubjectRepository


base_url: str = "https://cbra-oauth2.localhost"


@pytest.fixture
def alice() -> ISubject:
    return ref.subject.Subject(sub="alice@example.unimatrixone.io")


@pytest.fixture
def bob() -> ISubject:
    return ref.subject.Subject(sub="bob@example.unimatrixone.io")


@pytest.fixture
def trudy() -> ISubject:
    """An intruder."""
    return ref.subject.Subject(sub="trudy@example.unimatrixone.io")


@pytest.fixture
def current_subject(bob: ISubject) -> ISubject:
    return bob


@pytest.fixture
def users(alice: ISubject, bob: ISubject) -> dict[str, ISubject]:
    return {
        typing.cast(str, alice.sub): alice,
        typing.cast(str, bob.sub): bob
    }


@pytest_asyncio.fixture # type: ignore
async def keys() -> typing.Dict[str, typing.Dict[str, typing.Any]]:
    return {
        ref.DEFAULT_ENCRYPTION_KEY: {
            'provider': 'local',
            'kty': "OKP",
            'alg': 'EdDSA',
            'crv': 'Ed448',
            'use': 'sig',
            'tags': ['oauth2', 'server']
        },
        ref.DEFAULT_SIGNING_KEY: {
            'provider': 'local',
            'kty': "OKP",
            'alg': 'EdDSA',
            'crv': 'Ed448',
            'use': 'sig',
            'tags': ['oauth2', 'server']
        },
    }


@pytest.fixture
def keychain(keys: dict[str, typing.Any]) -> Keychain:
    keychain = Keychain()
    keychain.configure(keys)
    return keychain


@pytest_asyncio.fixture # type: ignore
async def codec(current_subject: ISubject) -> PayloadCodec:
    return await get_subject_codec(current_subject)


@pytest.fixture
def subjects(
    users: typing.Dict[str, ISubject]
) -> typing.Generator[typing.Type[ISubjectRepository], None, None]:
    MemorySubjectRepository.configure(users)
    yield MemorySubjectRepository
    MemorySubjectRepository.clear()


@pytest.fixture
def handlers():
    return [handle_rfc9068_token]


@pytest.fixture
def server(
    keys: typing.Dict[str, typing.Dict[str, typing.Any]],
    subjects: typing.Type[ISubjectRepository]
) -> AuthorizationServer:
    server = AuthorizationServer(
        authorize=AuthorizationEndpoint.new(
        ),
        storage=MemoryStorage,
        clients=ConfigFileClientRepository,
        grant_types={
            "authorization_code",
            "client_credentials",
            "urn:ietf:params:oauth:grant-type:jwt-bearer"
        },
        signing_key=ref.DEFAULT_SIGNING_KEY,
        subjects=subjects,
        principal_factory=cbra.auth.DebugPrincipal
    )
    return server


@pytest.fixture
def server_codec(server: AuthorizationServer):
    return PayloadCodec(
        decrypter=server.keychain,
        signer=server.keychain
    )


@pytest_asyncio.fixture # type: ignore
async def client(app: cbra.Application):
    params = {
        'app': app,
        'base_url': base_url
    }
    async with AsyncClient(**params) as client:
        yield client


async def handle_rfc9068_token(
    principal: IPrincipal = fastapi.Depends(RFC9068Principal())
):
    assert principal.has_scope("read")
    return {"message": "Hello world!"}


async def get_subject_codec(subject: ISubject) -> PayloadCodec:
    keychain = Keychain()
    keychain.configure({
        'sig': {
            'provider': 'local',
            'kty': 'OKP',
            'alg': 'EdDSA',
            'use': 'sig',
            'crv': 'Ed25519'
        }
    })
    await keychain
    private = keychain.get('sig')
    jwk = private.as_jwk(private=False)
    subject.register_public_key(jwk)
    # TODO: End hack

    return PayloadCodec(signer=keychain)


"""Declares :class:`SpaceSeparatedList`."""
import re
import typing


class SpaceSeparatedList(set[str]):
    __module__: str = 'cbra.ext.oauth2.types'
    sep: re.Pattern[str] = re.compile('\\s+')

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(
        cls,
        data: typing.Union[str, typing.List[str]]
    ) -> 'SpaceSeparatedList':
        if isinstance(data, (list, set)):
            return cls(data)
        return cls([x for x in cls.sep.split(data)]) if data else cls()

    @classmethod
    def __modify_schema__(
        cls,
        field_schema: typing.Dict[str, typing.Any]
    ):
        field_schema.update(
            type="string",
            example="value1 value2 value3"
        )

    def __repr__(self) -> str:
        return set(self).__repr__()

    def __str__(self) -> str:
        return str.join(' ', sorted(self))


# pylint: skip-file
# type: ignore
import pytest

from cbra.ext import oauth2


def test_redirect_uri_must_be_https():
    with pytest.raises(oauth2.exceptions.RedirectForbidden):
        oauth2.types.RedirectURL.validate("http://example.com")


def test_redirect_uri_must_have_netloc():
    with pytest.raises(oauth2.exceptions.RedirectForbidden):
        oauth2.types.RedirectURL.validate("http:///foo")


def test_redirect_uri_can_be_none():
    url = oauth2.types.RedirectURL.validate(None)
    assert not url


def test_redirect_uri_is_none_is_casted_to_empty_string():
    url = oauth2.types.RedirectURL.validate(None)
    assert str(url) == ''


def test_redirect_uri_is_none_raises_missing_redirect_wtihout_default():
    url = oauth2.types.RedirectURL.validate(None)
    with pytest.raises(oauth2.exceptions.MissingRedirectURL):
        url.authorize()


def test_redirect_uri_discovers_params():
    url = oauth2.types.RedirectURL.validate("https://example.com?foo=baz")
    assert url.params


def test_redirect_uri_authorize_adds_param():
    url = oauth2.types.RedirectURL.validate("https://example.com")
    assert url.authorize(foo=1) == "https://example.com?foo=1"


def test_redirect_uri_authorize_adds_param_array():
    url = oauth2.types.RedirectURL.validate("https://example.com")
    assert url.authorize(foo=[1,2]) == "https://example.com?foo=1&foo=2"


def test_redirect_uri_authorize_retains_param():
    url = oauth2.types.RedirectURL.validate("https://example.com?bar=1")
    assert url.authorize() == "https://example.com?bar=1"


def test_redirect_uri_authorize_retains_param():
    url = oauth2.types.RedirectURL.validate("https://example.com?bar=1&bar=2")
    assert url.authorize() == "https://example.com?bar=1&bar=2"


def test_redirect_uri_authorize_overwrites_param():
    url = oauth2.types.RedirectURL.validate("https://example.com?bar=1")
    assert url.authorize(bar=2) == "https://example.com?bar=2"


def test_redirect_uri_valid():
    url = oauth2.types.RedirectURL.validate("https://example.com")
    assert url


def test_redirect_uri_valid_to_string():
    url = oauth2.types.RedirectURL.validate("https://example.com")
    assert str(url) == "https://example.com"


def test_redirect_uri_raises_redirectforbidden_with_no_netloc():
    with pytest.raises(oauth2.exceptions.RedirectForbidden):
        url = oauth2.types.RedirectURL.validate("https:/example.com")


"""Declares :class:`ContentNegotiation`."""
import ckms.jose
import fastapi

import cbra
from cbra.headers import ACCEPT
#from cbra.headers import CONTENT_ENCODING
#from cbra.headers import CONTENT_LENGTH
from cbra.headers import CONTENT_TYPE
from cbra.headers import WANTS_DIGEST
from cbra.negotiation import DefaultContentNegotiation
from .params import ClientRepository
from .params import LocalIssuer
from .params import ServerCodec
from .types import IClientRepository


class ContentNegotiation(DefaultContentNegotiation):
    __module__: str = 'cbra.ext.oauth2'
    clients: IClientRepository
    codec: ckms.jose.PayloadCodec
    issuer: str

    @classmethod
    def with_body( # type: ignore
        cls,
        request: fastapi.Request,
        accept: str = ACCEPT,
        #content_encoding: str = CONTENT_ENCODING,
        #content_length: int = CONTENT_LENGTH,
        content_type: str = CONTENT_TYPE,
        wants_digest: str | None = WANTS_DIGEST,
        clients: IClientRepository = ClientRepository,
        codec: ckms.jose.PayloadCodec = ServerCodec,
        issuer: str = LocalIssuer
    ) -> cbra.types.IContentNegotiation:
        instance = cls(
            request=request,
            accept=accept,
            wants_digest=wants_digest
        )
        #instance.content_encoding = content_encoding
        #instance.content_length = content_length
        instance.content_type = content_type
        instance.clients = clients
        instance.codec = codec
        instance.issuer = issuer
        return instance

    def select_parser(
        self, parsers: list[type[cbra.types.IParser]]
    ) -> cbra.types.IParser:
        parser = super().select_parser(parsers)
        parser.server_codec = self.codec # type: ignore
        parser.clients = self.clients # type: ignore
        parser.issuer = self.issuer # type: ignore
        return parser


"""Declares :class:`MessagePublished`."""
import pydantic

from .pubsubmessage import PubsubMessage

SUBSCRIPTION_NAME = pydantic.constr(
    regex='^projects/[\-a-z0-9]{6,30}/subscriptions/.*$'
)

class MessagePublished(pydantic.BaseModel):
    """A datastructure containing a message that was published to a Google Pub/Sub
    topic and received by the server through a subscription.
    """
    subscription: SUBSCRIPTION_NAME = pydantic.Field(
        default=...,
        title="Subscription",
        description=(
            "The subscription through which the message was delivered to the "
            "endpoint."
        ),
        example="projects/myproject/subscriptions/mysub"
    )

    message: PubsubMessage



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any

import fastapi
from httpx import AsyncClient
from httpx import Response
from ckms.types import AuthorizationServerNotDiscoverable
from ckms.types import AuthorizationServerMisbehaves
from ckms.types import JSONWebKeySet
from ckms.types import ServerMetadata

from cbra.conf import settings
from cbra.utils import retry
from .credential import Credential
from .resource import Resource
from .serviceidentity import ServiceIdentity


RESOURCE_SERVERS: dict[str, Any] = getattr(settings, 'RESOURCE_SERVERS', {})


class ServiceClient:
    __module__: str = 'cbra.ext.service'
    credential: Credential | None
    http: AsyncClient
    identity: ServiceIdentity
    jwks: JSONWebKeySet | None
    logger: logging.Logger
    metadata: ServerMetadata
    server: str
    timeout: float
    resources: dict[str, Resource]

    @staticmethod
    def current(request: fastapi.Request) -> 'ServiceClient':
        return request.app.client

    def __init__(
        self,
        server: str,
        identity: ServiceIdentity,
        logger: logging.Logger | None = None,
        timeout: float = 60.0
    ):
        self.identity = identity
        self.logger = logger or logging.getLogger("uvicorn")
        self.resources = {}
        self.server = server
        self.timeout = timeout

    async def configure(self, resource: str) -> Resource:
        """Configure the named resource `resource` and return a
        :class:`~cbra.ext.service.Resource` instance.
        """
        if resource not in RESOURCE_SERVERS:
            raise ValueError(f"Resource not configured: {resource}")
        if resource not in self.resources:
            url = RESOURCE_SERVERS[resource]['server']
            self.resources[resource] = Resource(
                server=url,
                credential=await self.get_credential(
                    scope=RESOURCE_SERVERS[resource].get('scope') or set(),
                    resource=url
                )
            )
            await self.resources[resource].connect()
            self.logger.info("Configured resource server %s", url)
        return self.resources[resource]

    async def get_client_assertion(self) -> str:
        """Encode a client assertion used to authenticate with the authorization
        server.
        """
        return await self.identity.get_client_assertion(
            token_endpoint=self.metadata.token_endpoint
        )

    async def get_credential(
        self,
        scope: set[str],
        resource: str | None = None
    ) -> Credential:
        """Obtain a credential for the given resource."""
        return await Credential.obtain(
            http=self.http,
            token_endpoint=self.metadata.token_endpoint,
            client_id=self.identity.client_id,
            assertion_factory=self.get_client_assertion,
            scope=scope,
            resource=resource
        )

    @retry(6, interval=5.0)
    async def get_server_jwks(self) -> JSONWebKeySet | None:
        return await self.metadata.get_jwks(self.http)

    @retry(6, interval=5.0)
    async def get_server_metadata(self) -> ServerMetadata:
        return await ServerMetadata.discover(self.http, self.server)

    async def on_boot(self):
        try:
            self.logger.info("Booting service client")
            self.http = await AsyncClient(timeout=self.timeout).__aenter__()
            self.metadata = await self.get_server_metadata()
            self.jwks = await self.get_server_jwks()
        except AuthorizationServerNotDiscoverable:
            self.logger.fatal("Unable to discover authorization server %s", self.server)
            raise
        self.logger.info("Succesfully discovered authorization server %s", self.server)
        if not self.metadata.token_endpoint:
            self.logger.critical(
                "The authorization server did not advertise a token endpoint."
            )
            raise AuthorizationServerMisbehaves

    async def introspect(self, access_token: str) -> None:
        """Introspect the given access token using the Token Introspection
        Endpoint of the configured authorization server.
        """
        if not self.credential:
            self.credential = await self.get_credential(
                scope={"oauth2.introspect"},
                resource=self.server
            )
        raise NotImplementedError

    async def on_teardown(self):
        self.logger.info("Teardown service client")
        for resource in self.resources.values():
            await resource.close()
        await self.http.__aexit__(None, None, None)

    async def request(
        self,
        method: str,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ):
        """Make a HTTP request to the specified resource and path."""
        if resource not in self.resources:
            await self.configure(resource)
        return await self.resources[resource].request(
            method=method,
            path=path,
            headers=headers,
            **kwargs
        )

    async def delete(
        self,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ) -> Response:
        """Perform a DELETE request to the specified resource."""
        return await self.request(
            method="DELETE",
            resource=resource,
            path=path,
            headers=headers,
            **kwargs
        )

    async def get(
        self,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ) -> Response:
        """Perform a POST request to the specified resource."""
        return await self.request(
            method="GET",
            resource=resource,
            path=path,
            headers=headers,
            **kwargs
        )

    async def post(
        self,
        resource: str,
        path: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any
    ) -> Response:
        """Perform a POST request to the specified resource."""
        return await self.request(
            resource=resource,
            method="POST",
            path=path,
            headers=headers,
            **kwargs
        )


# pylint: disable=W0221,W0223
"""Declares :class:`Injected`."""
import inspect
import typing
from typing import Any
from typing import AsyncGenerator

import fastapi

from .dependency import Dependency
from .provider import _default


NULL = object()


class Injected(Dependency): # pragma: no cover
    """A :class:`cbra.Dependency` implementation that uses the :mod:`ioc`
    module to provide a dependency.
    """
    __module__: str = 'cbra.ext.ioc'
    args: list[Any]
    default: Any
    invoke: bool = False
    kwargs: dict[str, Any]

    def __init__(
        self,
        name: str,
        invoke: bool = False,
        default: typing.Any = NULL
    ):
        self.name = name
        self.invoke = invoke
        self.default = default
        self.args = []
        self.kwargs = {}
        super().__init__(use_cache=True)

    def get_signature(self) -> inspect.Signature:
        """Return a :class:`inspect.Signature` instance representing the
        call sigature of the dependency resolver.
        """
        return inspect.signature(self.__call__)

    async def __call__(self, request: fastapi.Request) -> AsyncGenerator[Any, None]:
        provider = getattr(request.app, 'container', _default)
        if not provider.is_satisfied(self.name) and self.default != NULL:
            dependency = self.default
        else:
            dependency = provider.resolve(self.name)
        yield dependency\
            if not self.invoke or not callable(dependency)\
            else dependency(*self.args, **self.kwargs)




# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from ckms.core import Keychain
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken
from ckms.utils import b64encode_str

from cbra.conf import settings


class PersonalDataHandler:
    """Exposes an interface to handle personal data."""
    __module__: str = 'cbra.ext.service'
    encryption_key: str | None = getattr(settings, 'PII_ENCRYPTION_KEY', None)
    codec: PayloadCodec
    index_key: str | None = getattr(settings, 'PII_INDEX_KEY', None)
    keychain: Keychain

    def __init__(self, keychain: Keychain):
        self.codec = PayloadCodec(
            decrypter=keychain,
            encrypter=keychain,
        )
        self.keychain = keychain

    async def decrypt(self, token: str) -> dict[str, Any]:
        """Decrypt the token using JSON Web Encryption (JWT)."""
        jwt = await self.codec.decode(token)
        if not isinstance(jwt, JSONWebToken):
            raise TypeError("Expected the token to contain an encrypted JWT.")
        return jwt.dict()

    async def encrypt(self, **kwargs: Any) -> str:
        """Encrypt the keyword parameters using JSON Web Encryption (JWT)."""
        if self.encryption_key is None:
            raise ValueError(
                'Specify the indexing key with setting PII_ENCRYPTION_KEY'
            )
        return await self.codec.encode(
            payload=kwargs,
            encrypters=[self.encryption_key]
        )

    async def index(self, value: str, encoding: str = 'utf-8') -> str:
        """Create a HMAC of the given value to store it as an index
        for sensitive data. Return the ASCII-encoded digest.
        """
        if self.index_key is None:
            raise ValueError('Specify the indexing key with setting PII_INDEX_KEY')
        k = self.keychain.get(self.index_key)
        return b64encode_str(await k.sign(str.encode(value, encoding=encoding)))


"""Declares :class:`TokenRequestHandler`."""
import typing

import fastapi
from ckms.types import WrongAudience
from unimatrix.exceptions import CanonicalException

from .endpoint import Endpoint
from .exceptions import CrossOriginNotAllowed
from .exceptions import Error
from .exceptions import InvalidGrant
from .params import Server
from .tokenissuer import TokenIssuer
from .tokenrequest import TokenRequest
from .types import BaseGrant
from .types import IAuthorizationServer
from .types import TokenResponse


class TokenRequestHandler(Endpoint):
    __module__: str = 'cbra.ext.oauth2'
    response_model: typing.Type[TokenResponse] = TokenResponse
    methods: set[str] = {"POST"}
    name: str = 'oauth2.token'
    summary: str = "Token Endpoint"
    description: str = (
        "The **Token Endpoint** is used by an application in order to "
        "get an access token or a refresh token. It is used by all "
        "flows except for the *Implicit Flow* because in that case an "
        "access token is issued directly.\n\n"
        "Depending on the client, the `Authorization` header or "
        "the `client_id` and `client_secret` body parameters may be "
        "mandatory to authenticate a confidential client."
    )

    #: A :class:`~cbra.ext.oauth2.TokenIssuer` instance that is used
    #: to issue tokens in response to requests to the **Token Endpoint**.
    issuer: TokenIssuer

    #: Description for the OPTIONS request handler.
    options_description: str = (
        "Communicates the allowed methods and CORS options for "
        "the **Token Endpoint**."
    )

    #: The parameters provided with the token request.
    params: BaseGrant

    def __init__(
        self,
        issuer: TokenIssuer = fastapi.Depends(),
        params: BaseGrant = TokenRequest
    ):
        self.issuer = issuer
        self.params = params

    async def enforce_cors_policy(
        self,
        origin: str | None = fastapi.Header(
            default=None,
            alias='Origin'
        )
    ):
        client = self.params.get_client()
        if origin and not client.allows_origin(origin):
            raise CrossOriginNotAllowed(model='client')

    async def handle(
        self,
        server: IAuthorizationServer = Server
    ) -> dict[str, typing.Any]:
        """Invokes the :attr:`TokenRequestHander.issuer` with the
        given grant type and return the result to the client.
        """
        if not server.allows_grant(self.params.grant_type.value):
            raise InvalidGrant(
                error_description=(
                    "The server does not allow use of the grant \""
                    f"{self.params.grant_type}\"."
                ),
                mode='client'
            )
        return (await self.issuer.grant(self.params))\
            .dict(exclude_defaults=True, exclude_none=True)

    async def on_exception( # type: ignore
        self,
        exception: Exception
    ) -> fastapi.Response:
        if isinstance(exception, WrongAudience):
            raise Error(
                error="invalid_grant",
                error_description=(
                    "The audience of the supplied JWT was not accepted "
                    "by the server."
                )
            )
        elif isinstance(exception, CanonicalException):
            raise Error(
                error="server_error",
                error_description=exception.message
            )
        raise


# pylint: skip-file
from typing import AsyncGenerator

import google.oauth2.id_token
import google.auth.transport.requests
import pytest
import pytest_asyncio
from httpx import AsyncClient

from cbra import Application
from examples.google import get_asgi_application


def get_id_token(audience: str) -> str:
    request = google.auth.transport.requests.Request()
    id_token = google.oauth2.id_token.fetch_id_token( # type: ignore
        request=request,
        audience=audience
    )
    return str(id_token) # type: ignore


@pytest.fixture(scope='session')
def id_token(base_url: str) -> str:
    return get_id_token(base_url)


@pytest_asyncio.fixture # type: ignore
async def app() -> Application:
    app = get_asgi_application()
    await app.on_startup()
    return app


@pytest_asyncio.fixture # type: ignore
async def client(
    app: Application,
    base_url: str
) -> AsyncGenerator[AsyncClient, None]:
    params = {
        'app': app,
        'base_url': base_url
    }
    async with AsyncClient(**params) as client:
        yield client



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`EventarcEndpoint`."""
from typing import Any

from cbra.negotiation import NullResponseContentNegotiation
from cbra.parsers import JSONParser
from cbra.responses import EmptyResponse
from cbra.types import IContentNegotiation
from cbra.types import IParser
from cbra.types import IRenderer
from cbra.utils import classproperty
import fastapi
from .googleendpoint import GoogleEndpoint
from .models import MessagePublished
from .models import PubsubMessage


class EventarcEndpoint(GoogleEndpoint):
    """A :class:`~cbra.ext.google.GoogleEndpoint` implementation that handles
    messages pushed by Google Eventarc.
    """
    __module__: str = 'cbra.ext.google'
    response_class: type[EmptyResponse] = EmptyResponse
    default_response_code: int = 200
    model: type[MessagePublished] = MessagePublished
    negotiation: type[IContentNegotiation] = NullResponseContentNegotiation
    parsers: list[type[IParser]] = [JSONParser]
    renderers: list[type[IRenderer]] = []
    response_description: str = "The message is succesfully processed."
    summary: str = "Google Eventarc"
    tags: list[str] = ["Cloud Events"]

    @classproperty
    def responses(cls) -> dict[str | int, Any]:
        """The responses returned by this endpoint."""
        defaults = super().responses
        defaults.update({
            202: {
                'headers': cls.common_headers,
                'description': (
                    "The message processing or queued for processing."
                )
            }
        })
        return defaults

    async def handle(self, dto: MessagePublished) -> None | fastapi.Response: # type: ignore
        """Receive a `google.cloud.pubsub.topic.v1.messagePublished` message
        and invoke the appropriate handler.
        """
        self.logger.debug(
            "Received message %s from Google Pub/Sub",
            dto.message.message_id
        )
        return await self.on_message(dto.message)

    async def on_message(self, message: PubsubMessage) -> None | fastapi.Response:
        """Handles the message received from Google Pub/Sub. The default
        implementation raises a :exc:`NotImplementedError`.
        """
        raise NotImplementedError


"""Declares common OAuth 2.0 exceptions."""
import functools
import typing
import urllib.parse
from typing import cast
from typing import Any

import fastapi
import fastapi.responses
import pydantic
from unimatrix.exceptions import CanonicalException
from unimatrix.lib.http import parse_qs

from cbra.exceptions import BaseRedirectingException


__all__ = [
    'ClientAuthenticationFailed',
    'Error',
    'ErrorModel',
    'InvalidClient',
    'LoginRequired',
    'MissingRedirectURL',
    'RedirectForbidden',
    'SubjectDoesNotExist',
    'TokenExchangeTargetRejected',
    'UnsupportedTokenTypeRequested',
    'UnsupportedResponseType',
    'UnsupportedSubjectTokenType',
]


class LoginRequired(BaseRedirectingException):
    """Raised when the resource owner needs to authenticate with
    the authorization server.
    """
    __module__: str = 'cbra.ext.oauth2.exceptions'


class ErrorModel(pydantic.BaseModel):
    error: str
    error_description: typing.Optional[str] = None
    error_uri: typing.Optional[str] = None


class Error(Exception):
    error: str = "invalid_request"
    error_description: str
    mode: str = 'client'
    state: str | None = None

    @classmethod
    def as_redirect(cls, redirect_uri: typing.Any) -> 'Error':
        return cls(redirect_uri=redirect_uri)

    @classmethod
    def catch(
        cls,
        error: typing.Type[BaseException],
        description: typing.Optional[str] = None
    ) -> typing.Callable[..., typing.Any]:
        def decorator_factory(func: typing.Any) -> typing.Callable[..., typing.Any]:
            @functools.wraps(func)
            async def decorated(*args: typing.Any, **kwargs: typing.Any):
                try:
                    return await func(*args, **kwargs)
                except error as exc:
                    msg = description
                    if isinstance(exc, CanonicalException) and not msg:
                        msg = exc.message or exc.code
                    if callable(description):
                        msg = typing.cast(str, description(exc))
                    raise cls(error_description=msg)
            return decorated
        return decorator_factory

    def __init__(self, **kwargs: str | typing.Any):
        kwargs.setdefault('error', self.error)
        kwargs.setdefault('error_description', self.get_error_description(**kwargs))
        self.mode = kwargs.pop('mode', self.mode)
        self.state = kwargs.pop('state', None)
        self._redirect_uri = kwargs.pop('redirect_uri', None)
        self._error = ErrorModel(
            error=typing.cast(str, kwargs['error']),
            error_description=typing.cast(str, kwargs['error_description'])
        )

    async def as_response(
        self,
        request: fastapi.Request,
        mode: str | None = None,
        state: str | None = None,
        error_url: str | None = None
    ) -> fastapi.responses.JSONResponse | fastapi.responses.RedirectResponse:
        """Return a :class:`fastapi.responses.Response` instance
        representing the error.
        """
        state = state or self.state
        mode = mode or self.mode
        if mode == 'client' or not self._redirect_uri:
            return fastapi.responses.JSONResponse(
                self._error.dict(exclude_defaults=True),
                status_code=400
            )
        else:
            params = {
                'error': self._error.error,
                'error_description': self._error.error_description
            }
            if state is not None:
                params['state'] = state
            return fastapi.responses.RedirectResponse(
                self._redirect_uri.authorize(**params),
                status_code=303
            )

    def dict(self, *args: Any, **kwargs: Any) -> dict[str, Any]:
        return {'error': self.error, 'error_description': self.error_description}

    def get_error_description(self, **kwargs: str) -> str:
        """Return the error description."""
        return kwargs.get('error_description') or self.error_description

    def must_redirect(self) -> bool:
        return self.mode == 'redirect'


class ClientAuthenticationFailed(Error):
    """Raised when the client credentials could not be authenticated."""
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_request"
    error_description: str = (
        "The client could not be authenticated using the given credentials."
    )


class InvalidGrant(Error):
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_grant"


class AssertionReplayed(InvalidGrant):
    """Raised when the credential is not valid because its replayed."""
    __module__: str = 'cbra.ext.oauth2'
    error_description: str = (
        "The assertion provided by the client was used previously to "
        "obtain an access token. Do not reuse assertions."
    )


class RedirectForbidden(Error):
    """Raised when the ``redirect_uri`` parameter is not valid
    for the given client.
    """
    error: str = "invalid_request"
    error_description: str = (
        "The client does not allow redirects to the URL supplied as "
        "the redirect_uri query parameter."
    )


class MissingRedirectURL(Error):
    error: str = "invalid_request"
    error_description: str = (
        "The `redirect_uri` parameter was not provided and the client "
        "did not specify a default."
    )


class RequestNotSupported(Error):
    error: str = "request_not_supported"
    error_description: str = (
        "The server does not support the use of the `request` or "
        "`request_uri` parameters."
    )


class InvalidClient(Error):
    error: str = "invalid_request"
    error_description: str = (
        "Specify the `client_id` parameter in the request body "
        "if the client is public, else provide credentials with "
        "the `Authorization` header."
    )


class InvalidScope(Error):
    error: str = "invalid_scope"
    error_description: str = (
        "The requested scope exceeds the scope that was granted by the "
        "resource owner."
    )


class ForbiddenScope(Error):
    error: str = "invalid_scope"
    error_description: str = (
        "The client application is not allowed or not allowing use of "
        "the requested scope."
    )


class ClientDoesNotExist(InvalidClient):
    """Like :exc:`InvalidClient`, but explicitely indicates that the
    client does not exist.
    """
    __module__: str = 'cbra.ext.oauth2'
    error_description: str = "The specified client does not exist."


class ClientAuthenticationRequired(InvalidClient):
    """Raised when there is an issued with authenticating a condfidential
    OAuth 2.0 client."""
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_client"
    error_description: str = "The specified client requires authentication."
    mode: str = 'client'


class InvalidClientAuthenticationScheme(ClientAuthenticationRequired):
    """Raised when the client used the wrong authentication scheme."""
    __module__: str = 'cbra.ext.oauth2'

    def __init__(self, using: str, required: str):
        super().__init__(
            error_description=(
                "Invalid authentication scheme. The server supports "
                f"{required} authentication for confidential clients."
            )
        )


class NoSuchClient(InvalidClient):
    """Raised during a request to the Token Endpoint when the client
    could not be determined from any factor.
    """
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_client"
    error_description: str = (
        "The client specified by the request parameters does not exist."
    )


class UnsupportedTokenTypeRequested(Error):
    """Raised when an unsupported token type is requested."""

    def __init__(self, requested: str, supported: list, **kwargs):
        Error.__init__(
            self,
            error="invalid_request",
            error_description=(
                "The server is unable to issue a token of the requested type: "
                f"{requested}. Supported token types are: "
                f"{str.join(',', sorted(supported))}."
            ),
            **kwargs
        )


class UnsupportedSubjectTokenType(Error):
    """Raised when an unsupported token type is provided."""

    def __init__(self, given: str, supported: list, **kwargs):
        Error.__init__(
            self,
            error="invalid_request",
            error_description=(
                "The server is unable to process a token of the provided type: "
                f"{given}. Supported token types are: "
                f"{str.join(',', sorted(supported))}."
            ),
            **kwargs
        )


class SubjectDoesNotExist(Error):
    """Raises when a lookup for a :term:`Subject` is attempted
    but the search query did not produce a result.
    """
    __module__: str = 'cbra.ext.oauth2'
    error: str = "invalid_request"
    error_description: str = (
        "The principal attached to the request did not resolve "
        "to a subject."
    )


class TokenExchangeTargetRejected(Error):
    """Raised when the server does not allow a specific audience or resource that
    is requested during a token exchange.
    """

    def __init__(self, target: str, message: typing.Optional[str] = None, **kwargs):
        Error.__init__(
            self,
            error="invalid_target",
            error_description=message or (
                f"The target {target} rejected the claims, scope or subject of "
                "the token exchange."
            ),
            **kwargs
        )


class UnsupportedResponseType(Error):
    __module__: str = 'cbra.ext.oauth2'
    error: str = (
        "The requested response type is not supported by the "
        "client or server."
    )
    error_description: str = (
        "Consult the documentation for the supported "
        "response types."
    )
    mode: str = 'redirect'


class UnsupportedResponseMode(Error):
    __module__: str = 'cbra.ext.oauth2'
    error: str = (
        "The requested response mode is not supported by the server."
    )
    error_description: str = (
        "Consult the documentation for the supported "
        "response types."
    )
    mode: str = 'redirect'


class ClientUnsupportedResponseMode(Error):
    __module__: str = 'cbra.ext.oauth2'
    error: str = (
        "The requested response mode is not allowed by the "
        "client."
    )
    error_description: str = (
        "Consult the documentation for the supported "
        "response types."
    )
    mode: str = 'redirect'


class CrossOriginNotAllowed(Error):
    error: str = "invalid_origin"
    error_description: str = (
        "Security measures prevent the authorization server from "
        "responding to requests from this domain, because it has "
        "not been whitelisted by the system administrator for use "
        "with this application."
    )
    mode: str = 'redirect'


# pylint: skip-file
import httpx
import pytest_asyncio


@pytest_asyncio.fixture
async def client(app):
    params = {
        'base_url': "https://cbra.unimatrixone.localhost",
        'app': app
    }
    async with httpx.AsyncClient(**params) as client:
        yield client


# pylint: skip-file
import typing

import httpx
from ckms.jose import PayloadCodec
from ckms.types import JSONWebToken


class JARMixin:
    base_url: str
    client_id: str
    jar_url: str

    async def create_jar_response(
        self,
        *,
        params: dict[str, typing.Any],
        codec: PayloadCodec,
        client: httpx.AsyncClient,
        algorithm: str = 'EdDSA',
        using: str = 'sig',
        headers: dict[str, str] | None = None
    ) -> tuple[str, str]:
        """Create a request to the authorization endpoint."""
        response = await client.post( # type: ignore
            url=self.jar_url,
            content=await self.create_jar(
                codec=codec,
                params=params,
                algorithm=algorithm,
                using=using
            ),
            headers={
                'Content-Type': "application/oauth-authz-req+jwt",
                **(headers or {})
            }
        )
        response.raise_for_status()
        assert response.status_code in {200, 201}, response.text
        dto = response.json()
        assert 'request_uri' in dto
        assert 'expires_in' in dto
        assert isinstance(dto['expires_in'], int)
        return dto['request_uri'], dto['expires_in']

    async def create_jar(
        self,
        *,
        codec: PayloadCodec,
        params: dict[str, typing.Any],
        algorithm: str = 'EdDSA',
        using: str = 'sig'
    ) -> str:
        """Creates a JWT-Secured Authorization Request."""
        return await self.serialize_jar(
            codec=codec,
            payload=JSONWebToken.strict(**{
                'iss': self.client_id,
                'aud': self.base_url,
                'client_id': self.client_id,
                'ttl': 120,
                **params
            }),
            using=using
        )

    async def serialize_jar(
        self,
        codec: PayloadCodec,
        payload: JSONWebToken,
        using: str
    ) -> str:
        """Serializes the payload of the JWT-Secured Authorization
        Request.
        """
        return await codec.encode(
            payload,
            signers=[using],
            content_type="oauth-authz-req+jwt"
        )


"""Declares :class:`IAuthorizeEndpoint`."""
from typing import Any

from cbra.types import IPrincipal
from .authorizationrequest import AuthorizationRequest
from .iclient import IClient
from .istorage import IStorage
from .isubject import ISubject
from .isubjectrepository import ISubjectRepository


class IAuthorizeEndpoint:
    __module__: str = 'cbra.ext.oauth2.types'
    error_url: str | None
    queryable: bool = True
    redirect_uri: str | None
    client: IClient
    issuer: str
    storage: IStorage
    subjects: ISubjectRepository

    async def create_redirect(
        self,
        dto: Any
    ) -> str:
        raise NotImplementedError

    async def get_subject(
        self,
        client: IClient,
        principal: IPrincipal,
        dto: AuthorizationRequest
    ) -> ISubject | None:
        """Return a :class:`~cbra.ext.oauth2.types.ISubject` instance
        representing the currently authenticated principal.
        """
        raise NotImplementedError

    async def persist(
        self,
        dto: AuthorizationRequest
    ) -> None:
        """Inspect the request parameters and persist the appropriate
        objects to the transient storage.
        """
        raise NotImplementedError

    async def validate_request(
        self,
        client: IClient,
        subject: ISubject,
        dto: AuthorizationRequest
    ) -> None:
        """Validates an OAuth 2.0 authorization request."""
        raise NotImplementedError


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from ckms.jose import PayloadCodec
from ckms.types import IKeySpecification
from ckms.types import JSONWebToken
from ckms.types import JSONWebKeySet
from httpx import AsyncClient


@pytest.mark.asyncio
async def test_cors_allows_different_origin(
    client: AsyncClient,
    codec: PayloadCodec,
    reks_url: str
):
    response = await client.get( # type: ignore
        url=reks_url,
        headers={'Origin': 'www.example.com'}
    )
    assert response.status_code == 200, response.content


@pytest.mark.asyncio
async def test_verify_jwks(
    client: AsyncClient,
    codec: PayloadCodec,
    reks_url: str
):
    response = await client.get(reks_url) # type: ignore
    jwt = await codec.decode(response.content, accept={"jwt+reks"})
    assert isinstance(jwt, JSONWebToken)
    assert 'jwks' in jwt.extra


@pytest.mark.asyncio
async def test_post_encrypted_and_signed_body(
    client: AsyncClient,
    codec: PayloadCodec,
    reks_url: str,
    jose_client_signer: IKeySpecification
):
    # Obtain the Request Encryption Key Set (REKS) from the
    # well-known endpoint.
    response = await client.get(reks_url) # type: ignore
    _, jwt = await codec.jwt(response.content, accept={"jwt+reks"})
    jwks = JSONWebKeySet(**jwt.extra['jwks'])


    # Use the REKS to encrypt a JSON Web Signature (JWS) and make
    # a POST request.
    assert jwks.keys
    client_codec = PayloadCodec(
        signing_keys=[jose_client_signer],
        encryption_keys=[x for x in jwks.keys]
    )
    obj = await client_codec.encode(
        payload={'foo': 1}
    )
    response = await client.post( # type: ignore
        url='/',
        content=obj,
        headers={'Content-Type': "application/jwt"}
    )
    dto = response.json()
    assert response.status_code == 200, dto['spec']['message']
    assert dto.get('foo') == 1, dto


# pylint: skip-file
# type: ignore
import typing

import pytest
from httpx import AsyncClient

from cbra import Application
from cbra.ext.oauth2 import ClientConfig
from cbra.ext.oauth2.types import GrantType
from .rfc9068resource import RFC9068Resource
from .servertest import ServerTest
from .tokenmixin import TokenMixin


class JWTAssertionMixin(ServerTest, TokenMixin):
    url: str = '/resources/1'

    @pytest.fixture
    def clients_available(self):
        return {
            'jwt': ClientConfig(
                client_id='jwt',
                grant_types_supported=[
                    GrantType.jwt_bearer.value
                ]
            )
        }

    @pytest.fixture
    def handlers(self) -> typing.List[typing.Any]:
        return [(RFC9068Resource, '/')]


class TestAccessToken(JWTAssertionMixin):

    @pytest.mark.asyncio
    async def test_missing_token_allows_no_access(self, client: AsyncClient):
        response = await client.get(url=self.url)
        assert response.status_code == 401

    @pytest.mark.asyncio
    async def test_valid_token_allows_access(self, client: AsyncClient, token: str):
        response = await client.get(
            url=self.url,
            headers={'Authorization': f"Bearer {token}"}
        )
        assert response.status_code == 200, response.text


class TestAccessTokenWithoutScope(JWTAssertionMixin):
    default_scope = set()

    @pytest.mark.asyncio
    async def test_rejects_token(self, client: AsyncClient, token: str):
        response = await client.get(
            url=self.url,
            headers={'Authorization': f"Bearer {token}"}
        )
        assert response.status_code == 403, response.text


class TestAccessTokenDifferentScope(JWTAssertionMixin):
    default_scope = {"foo"}

    @pytest.mark.asyncio
    async def test_rejects_token(self, client: AsyncClient, token: str):
        response = await client.get(
            url=self.url,
            headers={'Authorization': f"Bearer {token}"}
        )
        assert response.status_code == 403, response.text



class TestAccessTokenExtraScope(JWTAssertionMixin):
    default_scope = {"foo", "read"}

    @pytest.mark.asyncio
    async def test_accepts_token(self, client: AsyncClient, token: str):
        response = await client.get(
            url=self.url,
            headers={'Authorization': f"Bearer {token}"}
        )
        assert response.status_code == 200, response.text



"""Declares :class:`PushedAuthorizationRequestEndpoint.`"""
import fastapi

import cbra
from .contentnegotiation import ContentNegotiation
from .exceptions import CrossOriginNotAllowed
from .endpoint import Endpoint
from .exceptions import Error
from .params import ClientRepository
from .params import CurrentServerMetadata
from .params import OIDCTokenBuilder
from .params import SubjectRepository
from .params import TransientStorage
from .parsers import JARParser
from .types import AuthorizationRequest
from .types import IClient
from .types import IClientRepository
from .types import IStorage
from .types import ISubject
from .types import ISubjectRepository
from .types import JAR
from .types import IOIDCTokenBuilder
from .types import ServerMetadata


class PushedAuthorizationRequestEndpoint(Endpoint):
    __module__: str = 'cbra.ext.oauth2'
    allowed_methods: set[str] = {"POST", "OPTIONS"}
    default_ttl: int = 3600
    client: IClient | None
    error_mode: str = 'client'
    storage: IStorage
    metadata: ServerMetadata = CurrentServerMetadata
    methods: set[str] = {"POST"}
    summary: str = "Pushed Authorization Request"
    description: str = (
        "Clients may create prepared authorization requests by pushing "
        "the parameters, as used with a regular authorization request, "
        "to this endpoint. The parameters are serialized as JSON object "
        "and subsequently encoded using one of the content types listed "
        "below.\n\n"
        "Payloads with the following content types are accepted by the "
        "Pushed Authorization Request endpoint:\n\n"
        "- `application/oauth-authz-req+jwt` - a JWT-Secured Authorization "
        "Request (JAR), as specified in RFC 9101. Only confidential clients "
        "are able to invoke the endpoint using this payload type, since the "
        "specification requires the JAR to be signed. After signing a JAR, "
        "it may be encrypted using any of the public keys published at the "
        "`jwks_uri` declared by the authorization servers' metadata.\n"
        "- `application/jwt` - the payload is encrypted using JSON Web "
        "Encryption (JWE) containing a JSON Web Token (JWT). It **must "
        "not** contain a JSON Web Signature (JWS). This payload type is "
        "only accepted for public clients. This a proprietary extension "
        "to the OAuth 2.0 specifications. Sending this payload using a "
        "confidential client results in an `invalid_client` error. The JWT "
        "must be encrypted using any of the encryption public keys published "
        "at the `jwks_uri` declared by the authorization servers' "
        "metadata.\n\n"
        "In addition to the parameters accepted by the **Authorization "
        "Endpoint**, the following parameters are defined for this "
        "endpoint:\n\n"
        "- `encryption_key` - the public key from an asymmetric keypair, "
        "represented in the JSON Web Key (JWK) format, that is used to "
        "encrypt the response from the **Token Endpoint**. The keypair "
        "is considered ephemeral - a new keypair **must** be generated "
        "for each pushed authorization request.\n"
        "- `id_token` - an OpenID Connect ID Token that was signed by "
        "an issuer trusted by the authorization server. The ID Token is "
        "encoded in a JSON Web Signature (JWS). This parameter may be "
        "used if the **Authorization Endpoint** has no means to identify "
        "the resource owner on whose behalf the authorization request is "
        "made. If it *does* know how to identify the resource owner, then "
        "`id_token` **should** be omitted, and in the case that it is not, "
        "the `sub` claim **must** match the subject identifier of the "
        "resource owner identified by the **Authorization Endpoint**.\n"
        "- `access_token` - like `id_token`, but an RFC 9068 access token "
        "signed by a trusted issuer.\n\n"
        "If either the of the above parameters are included in the "
        "authorization request, the request **must** be encrypted."
    )
    negotiation: type[cbra.types.IContentNegotiation] = ContentNegotiation
    renderers: list[type[cbra.types.IRenderer]] = [
        cbra.renderers.JSONRenderer
    ]
    openid: IOIDCTokenBuilder = OIDCTokenBuilder
    options_description: str = (
        "Communicates the allowed methods and CORS options for "
        "the **Pushed Authorization Request** endpoint."
    )
    parsers: list[type[cbra.types.IParser]] = [
        JARParser
    ]

    def __init__(
        self,
        clients: IClientRepository = ClientRepository,
        storage: IStorage = TransientStorage,
        subjects: ISubjectRepository = SubjectRepository
    ):
        self.client = None
        self.clients = clients
        self.storage = storage
        self.subjects = subjects

    async def enforce_cors_policy( # type: ignore
        self,
        dto: JAR,
        origin: str | None = fastapi.Header(
            default=None,
            alias='Origin'
        )
    ) -> None:
        """Determine if the client allows a cross-origin request from the
        given origin.
        """
        client = await self.get_client(dto)
        if not client.allows_origin(origin):
            raise CrossOriginNotAllowed

    def create_urn(self, dto: AuthorizationRequest) -> str:
        """Create the URN to be used in the ``request_uri`` parameter."""
        assert dto.request_id is not None # nosec
        return f'urn:ietf:params:oauth:request_uri:{dto.request_id}'

    async def get_client(self, par: JAR) -> IClient:
        """Return the client that is requesting access on the behalf
        of the resource owner.
        """
        if self.client is None:
            self.client = await self.clients.get(par.client_id)
        return self.client

    async def get_subject(self, par: JAR) -> ISubject | None:
        """Return the subject that represents the resource owner."""
        return None

    async def handle(self, par: JAR) -> dict[str, int | str]: # type: ignore
        """Handles a **Pushed Authorization Request (PAR)**. The signature,
        client and claims are considered to be validated by the request
        body parser.
        """
        dto = AuthorizationRequest.fromparams(par)
        try:
            await dto.validate_request_parameters(
                client=await self.get_client(par),
                subject=await self.get_subject(par),
                metadata=self.metadata,
                openid=self.openid
            )
        except Error as exception:
            # Set mode to 'client' to explicitely prevent redirects,
            # since this endpoint does not interact with the resource
            # owner.
            exception.mode = 'client'
            raise exception
        await self.storage.persist(dto)
        return {
            'request_uri': self.create_urn(dto),
            'expires_in': self.default_ttl
        }


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`ProductType`."""
import enum


class ProductType(str, enum.Enum):
    simple = "simple"
    grouped = "grouped"
    external = "external"
    variable = "variable"
    variation = "variation"


"""Declares :class:`CodeChallenge`."""
import os

import typing

import pydantic

from .codechallengemethod import CodeChallengeMethod
from .dependantmodel import DependantModel


CodeChallengeType = pydantic.constr(min_length=43, max_length=128)


class CodeChallenge(DependantModel):
    __module__: str = 'cbra.ext.oauth2.types'

    challenge: CodeChallengeType = pydantic.Field(
        default=None,
        title="Code challenge",
        alias='code_challenge',
        description=(
            'A high-entropy cryptographic random string using the unreserved '
            'characters [A-Z] / [a-z] / [0-9] / "-" / "." / "_" / "~" from '
            'Section 2.3 of RFC 3986, with a minimum length of 43 characters '
            'and a maximum length of 128 characters.'
        ),
        example=bytes.hex(os.urandom(56))
    )

    method: typing.Optional[CodeChallengeMethod] = pydantic.Field(
        default=None,
        title="Code challenge method",
        alias="code_challenge_method",
        description=(
            "Specifies the method used to calculate and verify the code "
            "challenge. Defaults to `plain` if not present in the request."
        ),
        enum=["plain", "S256"],
        example="S256"
    )


"""Declares :class:`MemoryClientRepository`."""
import typing

from .types import IClient
from .types import IClientRepository


class MemoryClientRepository(IClientRepository):
    """A :class:`IClientRepository` implementation that used local
    memory for storage. Mainly for testing purposes.
    """
    __module__: str = 'cbra.ext.oauth2'
    clients: typing.Dict[typing.Union[int, str], IClient] = {}

    @classmethod
    def add(cls, client: IClient) -> None:
        """Add a client to the repository."""
        cls.clients[client.client_id] = client

    @classmethod
    def clear(cls) -> None:
        """Clears the repository of all data."""
        cls.clients = {}

    @classmethod
    def configure(cls, clients: typing.Dict[str, IClient]) -> None:
        """Load a mapping of client identifiers to clients."""
        cls.clients.update(clients) # type: ignore

    def __init__(self):
        self.clients = MemoryClientRepository.clients

    async def exists(self, client_id: str) -> bool:
        return client_id in self.clients

    async def get(self, client_id: str) -> IClient:
        try:
            client = self.clients[client_id]
            return client
        except KeyError:
            raise self.ClientDoesNotExist


"""Declares :class:`NullSession`."""
from typing import Any

import fastapi

from .basesession import BaseSession


class NullSession(BaseSession):
    """A session implementation that does nothing."""
    __module__: str = 'cbra.session'

    async def add_to_response(self, response: fastapi.Response) -> None:
        pass

    async def get(self, key: str) -> Any:
        return None

    async def set(self, key: str, value: Any) -> None:
        pass


# pylint: skip-file
# type: ignore
import secrets

import pytest
from ckms.jose import PayloadCodec
from httpx import AsyncClient

from cbra.ext.oauth2 import ClientConfig
from cbra.ext.oauth2.types import GrantType
from cbra.ext.oauth2.types import ISubject
from cbra.ext.oauth2.types import TokenRequestParameters
from ...conftest import get_subject_codec
from ...servertest import ServerTest
from ...tokenmixin import TokenMixin


class TestJWTBearerAssertion(ServerTest, TokenMixin):

    async def get_token_request(self, **kwargs: str):
        defaults = {}
        defaults.update(**kwargs)
        return defaults

    def get_clients_available(self):
        return {
            'jwt': ClientConfig(
                client_id='jwt',
                grant_types_supported=[
                    GrantType.jwt_bearer.value
                ]
            )
        }

    @pytest.mark.asyncio
    async def test_token_endpoint(self, client: AsyncClient, grant: str):
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': grant
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        assert response.status_code == 200, response.text

    @pytest.mark.asyncio
    async def test_token_endpoint_with_disabled_scope(self, client: AsyncClient, grant: str):
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': grant,
            'scope': "openid profile foo"
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        self.assert_error(response, "invalid_scope")

    @pytest.mark.asyncio
    async def test_token_endpoint_with_non_consented_scope(self, client: AsyncClient, grant: str):
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': grant,
            'scope': "openid profile email"
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        self.assert_error(response, "invalid_scope")

    @pytest.mark.asyncio
    async def test_cannot_replay_token(self, client: AsyncClient, grant: str):
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': grant
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        assert response.status_code == 200, response.text
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        assert response.status_code == 400, response.text

    @pytest.mark.asyncio
    async def test_protected_endpoint_without_token(self, client: AsyncClient):
        response = await client.get('/rfc9068')
        assert response.status_code == 401, response.text

    @pytest.mark.asyncio
    async def test_protected_endpoint_with_valid_token(self, client: AsyncClient, token: str):
        response = await client.get(
            url='/rfc9068',
            headers={
                'Authorization': f"Bearer {token}"
            }
        )
        assert response.status_code == 200, response.text


    @pytest.mark.asyncio
    async def test_protected_endpoint_with_invalid_scheme(self, client: AsyncClient, token: str):
        response = await client.get(
            url='/rfc9068',
            headers={
                'Authorization': f"Foo {token}"
            }
        )
        assert response.status_code == 403, response.text

    @pytest.mark.asyncio
    async def test_token_endpoint_returns_invalid_client_on_non_existing(
        self,
        client: AsyncClient,
        grant: str
    ):
        request = dict(
            client_id=secrets.token_hex(2),
            grant_type=GrantType.jwt_bearer,
            assertion=grant
        )
        response = await client.post(url='/oauth2/token', json=request)
        self.assert_error(response, 'invalid_client')

    @pytest.mark.asyncio
    async def test_token_endpoint_returns_invalid_client_on_missing(
        self,
        client: AsyncClient,
        grant: str
    ):
        request = dict(
            grant_type=GrantType.jwt_bearer,
            assertion=grant
        )
        response = await client.post(url='/oauth2/token', json=request)
        self.assert_error(response, 'invalid_client')

    @pytest.mark.asyncio
    async def test_token_endpoint_rejects_assertion_with_signature_from_mismatching_issuer(
        self,
        client: AsyncClient,
        codec: PayloadCodec,
        current_subject: ISubject,
        alice: ISubject
    ):
        assert alice.sub != current_subject.sub
        # HACK
        await get_subject_codec(alice)
        grant = await self.sign_jwt(
            codec,
            iss=alice.sub,
            aud=f'{self.base_url}/oauth2/token',
            sub=alice.sub,
        )
        request = dict(
            client_id="jwt",
            grant_type=GrantType.jwt_bearer,
            assertion=grant
        )
        response = await client.post(url='/oauth2/token', json=request)
        self.assert_error(response, 'invalid_grant')

    @pytest.mark.asyncio
    async def test_token_endpoint_rejects_assertion_from_unknown_subject(
        self,
        client: AsyncClient,
        current_subject: ISubject,
        trudy: ISubject
    ):
        assert trudy.sub != current_subject.sub
        # HACK
        codec = await get_subject_codec(trudy)
        grant = await self.sign_jwt(
            codec,
            iss=trudy.sub,
            aud=f'{self.base_url}/oauth2/token',
            sub=trudy.sub,
        )
        request = dict(
            client_id="jwt",
            grant_type=GrantType.jwt_bearer,
            assertion=grant
        )
        response = await client.post(url='/oauth2/token', json=request)
        self.assert_error(response, 'invalid_request')

    @pytest.mark.asyncio
    async def test_token_endpoint_rejects_escalated_scoped(
        self,
        client: AsyncClient,
        grant: str
    ):
        request = dict(
            client_id="jwt",
            grant_type=GrantType.jwt_bearer,
            assertion=grant,
            scope=[secrets.token_hex(4)]
        )
        response = await client.post(url='/oauth2/token', json=request)
        self.assert_error(response, 'invalid_scope')

    @pytest.mark.asyncio
    async def test_token_endpoint_rejects_audience_other_than_itself(
        self,
        client: AsyncClient,
        codec: PayloadCodec,
        current_subject: ISubject,
    ):
        grant = await self.sign_jwt(
            codec,
            iss=current_subject.sub,
            aud=f'{self.base_url}',
            sub=current_subject.sub,
        )
        request = dict(
            client_id="jwt",
            grant_type=GrantType.jwt_bearer,
            assertion=grant,
            scope=[secrets.token_hex(4)]
        )
        response = await client.post(url='/oauth2/token', json=request)
        self.assert_error(response, 'invalid_grant')

    @pytest.mark.asyncio
    async def test_token_endpoint_rejects_token_audiences_that_is_not_approved(
        self,
        client: AsyncClient,
        grant: str
    ):
        request = dict(
            client_id="jwt",
            grant_type=GrantType.jwt_bearer,
            assertion=grant,
            resource="foo"
        )
        response = await client.post(url='/oauth2/token', json=request)
        self.assert_error(response, 'invalid_grant')

    @pytest.mark.asyncio
    async def test_malformed_returns_invalid_grant(self, client: AsyncClient, grant: str):
        header, payload, signature = str.split(grant, '.')
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': 'foo'
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        self.assert_error(response, "invalid_grant")

    @pytest.mark.asyncio
    async def test_malformed_header_returns_invalid_grant(self, client: AsyncClient, grant: str):
        header, payload, signature = str.split(grant, '.')
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': str.join('.', ['foo', payload, signature])
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        self.assert_error(response, "invalid_grant")

    @pytest.mark.asyncio
    async def test_malformed_payload_returns_invalid_grant(self, client: AsyncClient, grant: str):
        header, payload, signature = str.split(grant, '.')
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': str.join('.', [header, 'foo', signature])
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        self.assert_error(response, "invalid_grant")

    @pytest.mark.asyncio
    async def test_malformed_signature_returns_invalid_grant(self, client: AsyncClient, grant: str):
        header, payload, signature = str.split(grant, '.')
        request = TokenRequestParameters.parse_obj({
            'client_id': 'jwt',
            'grant_type': GrantType.jwt_bearer,
            'assertion': str.join('.', [header, payload, 'foo'])
        })
        response = await client.post(
            url='/oauth2/token',
            content=request.json(),
            headers={'Content-Type': "application/json"}
        )
        self.assert_error(response, "invalid_grant")


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`BaseParser`."""
import abc
from typing import Any

import fastapi

from cbra.types import IParser


class BaseParser(IParser, metaclass=abc.ABCMeta):
    """All parsers should extend `BaseParser`, specifying a `media_type`
    attribute, and overriding the `.parse()` method.
    """
    media_type: str

    @classmethod
    def openapi_example(cls, schema: dict[str, Any]) -> str:
        return ""

    async def parse(
        self,
        request: fastapi.Request,
        media_type: str | None = None,
        parser_context: dict[str, Any] | None = None
    ) -> Any:
        """ Given a stream to read from, return the parsed representation.
        Should return parsed data, or a `DataAndFiles` object consisting of the
        parsed data and files.
        """
        raise NotImplementedError(".parse() must be overridden.")


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestOrderUpdated(BaseMessageTest):
    event_type: str = 'updated'
    resource_type: str = 'order'

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': "application/json",
                'x-wc-webhook-delivery-id': "c3109c2183bf5e9b4d1e7d6dac13c3dd",
                'x-wc-webhook-event': "updated",
                'x-wc-webhook-id': "62",
                'x-wc-webhook-resource': "order",
                'x-wc-webhook-signature': "ltYxPLIY69upwj0XTf+TT6h7p20fCjEcItIbGTuyVyw=",
                'x-wc-webhook-source': "https://test-wordpress.molano.nl/",
                'x-wc-webhook-topic': "order.updated",
            },
            'content': (
                b'{"id":76,"parent_id":0,"status":"pending","currency":"EUR","version":"'
                b'5.9.1","prices_include_tax":false,"date_created":"2022-06-03T20:28:09"'
                b',"date_modified":"2022-06-03T20:45:55","discount_total":"0.00","discou'
                b'nt_tax":"0.00","shipping_total":"0.00","shipping_tax":"0.00","cart_tax'
                b'":"0.00","total":"0.00","total_tax":"0.00","customer_id":1,"order_key"'
                b':"wc_order_tpYsgYNGJe8OD","billing":{"first_name":"","last_name":"","c'
                b'ompany":"","address_1":"","address_2":"","city":"","state":"","postcod'
                b'e":"","country":"","email":"cochise.ruhulessin@unimatrixone.io","phone'
                b'":""},"shipping":{"first_name":"","last_name":"","company":"","address'
                b'_1":"","address_2":"","city":"","state":"","postcode":"","country":"",'
                b'"phone":""},"payment_method":"","payment_method_title":"","transaction'
                b'_id":"","customer_ip_address":"","customer_user_agent":"","created_via'
                b'":"admin","customer_note":"","date_completed":null,"date_paid":null,"c'
                b'art_hash":"","number":"76","meta_data":[],"line_items":[{"id":4,"name"'
                b':"Test Product p","product_id":39,"variation_id":0,"quantity":1,"tax_c'
                b'lass":"","subtotal":"0.00","subtotal_tax":"0.00","total":"0.00","total'
                b'_tax":"0.00","taxes":[],"meta_data":[],"sku":"XXXTEST","price":0,"pare'
                b'nt_name":null},{"id":5,"name":"variants 4 - foo, red","product_id":46,'
                b'"variation_id":57,"quantity":1,"tax_class":"","subtotal":"0.00","subto'
                b'tal_tax":"0.00","total":"0.00","total_tax":"0.00","taxes":[],"meta_dat'
                b'a":[{"id":49,"key":"teset","value":"foo","display_key":"teset","displa'
                b'y_value":"foo"},{"id":50,"key":"color","value":"red","display_key":"co'
                b'lor","display_value":"red"}],"sku":"XXXBAZ","price":0,"parent_name":"v'
                b'ariants 4"}],"tax_lines":[],"shipping_lines":[],"fee_lines":[],"coupon'
                b'_lines":[],"refunds":[],"date_created_gmt":"2022-06-03T20:28:09","date'
                b'_modified_gmt":"2022-06-03T20:45:55","date_completed_gmt":null,"date_p'
                b'aid_gmt":null,"currency_symbol":"\\u20ac","_links":{"self":[{"href":"ht'
                b'tps:\\/\\/test-wordpress.molano.nl\\/wp-json\\/wc\\/v3\\/orders\\/76"}],"coll'
                b'ection":[{"href":"https:\\/\\/test-wordpress.molano.nl\\/wp-json\\/wc\\/v3'
                b'\\/orders"}],"customer":[{"href":"https:\\/\\/test-wordpress.molano.nl\\/wp'
                b'-json\\/wc\\/v3\\/customers\\/1"}]}}'
            )
        }


"""Declares :class:`NullSubjectRepository`."""
from .types import ISubject
from .types import ISubjectRepository


class NullSubjectRepository(ISubjectRepository):
    __module__: str = 'cbra.ext.oauth2'

    async def get(
        self,
        subject_id: int | str,
        client_id: str | None = None
    ) -> ISubject:
        raise self.SubjectDoesNotExist


# pylint: skip-file
import asyncio
import datetime
import functools
import inspect
import itertools
import logging
import typing
from inspect import Parameter
from typing import Any, Coroutine
from typing import Callable

from .mediatypes import media_type_matches
from .mediatypes import parse_header


__all__ = [
	'media_type_matches',
    'normalize_signature',
    'parse_header',
    'retry',
]


class DeferredException:
    """Defer raising the exception until an attribute is accessed."""
    exc: BaseException

    def __init__(self, exc: BaseException):
        self.exc = exc

    def raise_exception(self) -> typing.NoReturn:
        raise self.exc

    def __getattr__(self, attname: str) -> typing.NoReturn:
        raise self.exc

    __bytes__ = raise_exception
    __str__ = raise_exception


def current_timestamp() -> int:
    return int(datetime.datetime.now(datetime.timezone.utc).timestamp())


def docstring(cls: type[typing.Any]): # type: ignore
    def update_func(func):
        func.__doc__ = getattr(cls, func.__name__).__doc__
        return func
    return update_func


def order_parameters(params: typing.List[Parameter]) -> typing.List[Parameter]:
    """Take a list of parameters and ensure that they are ordered correctly i.e.
    positional arguments, keyword arguments, variable positional arguments and
    variable keyword arguments.
    """
    default = [x for x in params if x.default != Parameter.empty]
    nondefault = [x for x in params if x.default == Parameter.empty]
    result = itertools.chain(
        nondefault,
        [x for x in default if x .kind == Parameter.POSITIONAL_ONLY],
        [x for x in default if x .kind == Parameter.POSITIONAL_OR_KEYWORD],
        [x for x in default if x .kind == Parameter.VAR_POSITIONAL],
        [x for x in default if x .kind == Parameter.KEYWORD_ONLY],
        [x for x in default if x .kind == Parameter.VAR_KEYWORD],
    )
    return list(result)


def normalize_signature(func):
    """Normalize the signature of a function or method so that it
    can be used as a :mod:`fastapi` request handler.
    """
    sig = inspect.signature(func)
    func.__signature__ = sig.replace(
        parameters=[
            x for x in sig.parameters.values()
            if x.name != 'self'
        ]
    )
    return func


class PositionalArgument(inspect.Parameter):

    def __init__(self,
        name: str,
        annotation: type = None,
        **kwargs
    ):
        super().__init__(
            name=name,
            kind=Parameter.POSITIONAL_ONLY,
            annotation=annotation,
            **kwargs
        )


class ClassPropertyDescriptor:

    @classmethod
    def new(cls, func: typing.Callable[..., typing.Any]) -> typing.Any:
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        return cls(func)

    def __init__(self, fget, fset=None):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj, cls=None) -> typing.Any:
        if cls is None:
            cls = type(obj)
        return self.fget.__get__(obj, cls)()

    def __set__(self, obj, value) -> typing.NoReturn:
        if not self.fset:
            raise AttributeError("can't set attribute")
        type_ = type(obj)
        return self.fset.__get__(obj, type_)(value)

    def setter(self, func) -> typing.Any:
        if not isinstance(func, (classmethod, staticmethod)):
            func = classmethod(func)
        self.fset = func
        return self


classproperty = ClassPropertyDescriptor.new


def retry(
    max_attempts: int,
    interval: float = 1.0
) -> Callable[..., Any]:
    """Decorator that retries the decorated function for given amount
    of `max_attempts.
    """
    logger = logging.getLogger("uvicorn")
    def decorator_factory(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        async def f(*args: Any, **kwargs: Any) -> Any:
            attempts: int = 0
            while True:
                try:
                    result = func(*args, **kwargs)
                    if inspect.isawaitable(result):
                        result = await result
                    break
                except Exception:
                    attempts += 1
                    if attempts >= max_attempts:
                        raise
                    logger.critical(
                        "Caught fatal exception while invoking %s(), retrying",
                        func.__name__
                    )
                    await asyncio.sleep(interval)
            return result
        return f
    return decorator_factory


"""Declares :class:`PathParameter`."""
import inspect
import typing

import fastapi


class PathParameter:
    __module__: str = 'cbra.resource'
    name: str | None
    annotation: typing.Type[typing.Any]
    path_annotations: typing.Dict[typing.Any, typing.Any] = {
        'int': int,
        int: 'int',
    }
    signature_parameter: inspect.Parameter | None

    def __init__(
        self,
        name: typing.Optional[str],
        annotation: typing.Type[typing.Any]
    ):
        if name is not None and ':' in name:
            self.name, kind = str.split(name, ':')
            if kind not in self.path_annotations:
                raise ValueError(f"Unknown filter: '{kind}'.")
            self.annotation = self.path_annotations[kind]
        else:
            self.name = name
            self.annotation = str
            self.signature_parameter = None
        if self.name is not None:
            self.signature_parameter = inspect.Parameter(
                name=self.name,
                kind=inspect.Parameter.POSITIONAL_OR_KEYWORD,
                annotation=self.annotation,
                default=fastapi.Path(...)
            )

    def insert_in(self, seq: list[inspect.Parameter]) -> None:
        if self.signature_parameter is not None:
            seq.insert(0, self.signature_parameter)

    def get_path_variable(self) -> str:
        assert self.name is not None # nosec
        name = self.name
        if self.annotation != str:
            name = f'{name}:{self.path_annotations[self.annotation]}'
        return self.name

    def get_path(
        self,
        *,
        base_path: str,
        subpath: typing.Optional[str],
        is_detail: bool
    ) -> str:
        path = base_path
        if self.name is not None and is_detail:
            path = f'{path}/{{{self.get_path_variable()}}}'
        if subpath:
            path = f'{path}/{subpath}'
        return path

    def __bool__(self) -> bool:
        return self.name is not None

    def __eq__(self, other: object) -> bool:
        other = typing.cast(PathParameter, other)
        return self.name == other.name


"""Declares :class:`UpstreamReturnHandler`."""
import fastapi
import pydantic

from .params import CurrentUpstreamProvider
from .params import TransientStorage
from .types import AuthorizationRequestParameters
from .types import IStorage
from .types import IUpstreamReturnHandler
from .types import IUpstreamProvider


class UpstreamReturnHandler(IUpstreamReturnHandler):
    __module__: str = 'cbra.ext.oauth2'

    @classmethod
    def add_to_router(
        cls,
        *,
        app: fastapi.FastAPI,
        base_path: str
    ) -> None:
        query_class = cls.query_model or pydantic.BaseModel

        async def handler(
            request: fastapi.Request,
            handler: cls = fastapi.Depends(),
            provider: IUpstreamProvider = CurrentUpstreamProvider(cls.provider),
            storage: IStorage = TransientStorage,
            params: query_class = fastapi.Depends()
        ) -> fastapi.Response:
            handler.storage = storage
            return await handler.handle(
                request=request,
                provider=provider,
                params=params
            )

        app.add_api_route(
            path=f'{base_path}/{cls.provider}',
            endpoint=handler,
            name=f'oauth2.callback.{cls.provider}',
            summary=cls.summary,
            description="Callback URL for external identity provider.",
            methods=['GET']
        )

    async def get_authorization_request(
        self,
        provider: IUpstreamProvider,
        request: fastapi.Request
    ) -> AuthorizationRequestParameters:
        return await provider.get_authorization_request(
            storage=self.storage,
            request=request
        )

    async def handle(
        self,
        *,
        request: fastapi.Request,
        provider: IUpstreamProvider,
        params: pydantic.BaseModel
    ) -> fastapi.Response:
        """Handles the redirect from the upstream identity provider and
        redirects the client to the URI specified in the authorization
        request.
        """
        authorize = await self.get_authorization_request(
            provider=provider,
            request=request
        )
        await provider.process_response(authorize, params)
        return fastapi.responses.RedirectResponse(
            url=provider.get_client_redirect(
                request=request,
                dto=authorize
            ),
            status_code=303
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from ckms.core import models


class JSONWebSignature(pydantic.BaseModel):
    payload: str = pydantic.Field(
        ...,
        title="Payload",
        description=(
            "The URL-safe, Base64-encoded `payload` member contains the "
            "UTF-8 JSON-serialized JSON Web Key Set (JWKS) holding the "
            "keys that may be used to encrypt HTTP requests to this "
            "server."
        )
    )

    signatures: list[models.Signature] = pydantic.Field(
        ...,
        title="Signatures",
        description=(
            "An array of `Signature` objects. Compliant implementations **must** "
            "verify each item in the `signatures` array."
        )
    )


"""Declares :class:`Options`."""
import typing

import fastapi

from cbra.exceptions import EXCEPTION_RESPONSES
from cbra.headers import ACCESS_CONTROL_ALLOW_CREDENTIALS_SCHEMA
from cbra.headers import ACCESS_CONTROL_ALLOW_HEADERS_SCHEMA
from cbra.headers import ACCESS_CONTROL_ALLOW_METHODS_SCHEMA
from cbra.headers import ACCESS_CONTROL_ALLOW_ORIGIN_SCHEMA
from cbra.headers import ACCESS_CONTROL_MAX_AGE_SCHEMA
from cbra.negotiation import NullResponseContentNegotiation
from cbra.responses import OptionsResponse
from cbra.types import IContentNegotiation
from cbra.types import IRenderer
from cbra.utils import classproperty
from .endpoint import Endpoint


class Options(Endpoint):
    __module__: str = 'cbra'
    allowed_methods: typing.Set[str] = set()
    description: str = (
        "Responds with the permitted communication options for a given "
        "URL or server. A client can specify a URL with this method, or "
        "an asterisk (*) to refer to the entire server. The response may "
        "include the following headers:\n\n"
        "- The `Allow` header, that indicates the allowed HTTP methods, "
        "separated by a comma.\n"
        "\n"
        "**Preflighted requests in CORS**\n\n"
        "In CORS, a preflight request is sent with the OPTIONS method so "
        "that the server can respond if it is acceptable to send the "
        "request. Permissions can be requested using HTTP headers for "
        "the following parameters:\n\n"
        "- The `Access-Control-Request-Method` header sent in a preflight "
        "request tells the server that when the actual request is sent, "
        "it will have the method specified as the header value.\n"
        "- The `Access-Control-Request-Headers` header tells the server "
        "that when the actual request is sent, it include the given headers. "
        "Multiple headers are separated by a comma.\n\n"
        "Depending on the server policy, it may respond with the following "
        "response headers:\n\n"
        "- The `Access-Control-Allow-Credentials` tells browsers whether to "
        "expose the response to the frontend JavaScript code when the request's "
        "credentials mode (`Request.credentials`) is `include`.\n"
        "- The `Access-Control-Request-Headers` to indicate which HTTP headers "
        "can be used during the actual request.\n"
        "- The `Access-Control-Allow-Methods` response header specifies one or "
        "more request methods allowed.\n"
        "- The `Access-Control-Allow-Origin` response header indicates whether "
        "the response can be shared with requesting code from the given origin."
    )
    document: bool = True
    method: str = "OPTIONS"
    negotiation: type[IContentNegotiation] = NullResponseContentNegotiation
    renderers: list[type[IRenderer]] = []
    responses: dict[int | str, typing.Any] = {
        500: EXCEPTION_RESPONSES[500]
    }
    response_class: type[OptionsResponse] = OptionsResponse

    @classproperty
    def default_response(cls) -> typing.Dict[str, typing.Any]:
        return {
            'description': "Endpoint allowed methods and CORS privileges.",
            'headers': {
                'Access-Control-Allow-Credentials': ACCESS_CONTROL_ALLOW_CREDENTIALS_SCHEMA,
                'Access-Control-Allow-Headers': ACCESS_CONTROL_ALLOW_HEADERS_SCHEMA,
                'Access-Control-Allow-Methods': ACCESS_CONTROL_ALLOW_METHODS_SCHEMA,
                'Access-Control-Allow-Origin': ACCESS_CONTROL_ALLOW_ORIGIN_SCHEMA,
                'Access-Control-Max-Age': ACCESS_CONTROL_MAX_AGE_SCHEMA
            }
        }

    async def handle(self) -> fastapi.Response:
        response = fastapi.Response()
        response.headers['Allow'] = str.join(', ', sorted(set(self.allowed_methods)|{"OPTIONS"}))\
            or "OPTIONS"
        return response



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""Declares :class:`ResourceLinks`."""
import pydantic


class Link(pydantic.BaseModel):
    href: str


class ResourceLinks(pydantic.BaseModel):
    """Metadata describing the endpoints where the resource(s) can be
    retrieved.
    """
    self: list[Link]
    collection: list[Link]



# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from .oidcclaimrequest import OIDCClaimRequest


Empty = type('Empty', (object,), {})


class OIDCRequestedClaims(pydantic.BaseModel):
    iss: OIDCClaimRequest | None = None
    sub: OIDCClaimRequest | None = None
    aud: OIDCClaimRequest | None = None
    exp: OIDCClaimRequest | None = None
    iat: OIDCClaimRequest | None = None
    auth_time: OIDCClaimRequest | None = None
    nonce: OIDCClaimRequest | None = None
    acr: OIDCClaimRequest | None = None
    amr: OIDCClaimRequest | None = None
    azp: OIDCClaimRequest | None = None


# Copyright (C) 2022 Cochise Ruhulessin <cochiseruhulessin@gmail.com>
# 
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
"""Declares :class:`NullCorsPolicy`."""
from typing import Any

import fastapi
import fastapi.params

from .basecorspolicy import BaseCorsPolicy


class NullCorsPolicy(BaseCorsPolicy):
    """Disable the CORS policy enforcement completely."""
    __module__: str = 'cbra.cors'
    origin_header: fastapi.params.Depends = fastapi.Depends(lambda: None)

    @classmethod
    def as_options(cls) -> 'ICorsPolicy': # type: ignore
        return cls()

    @classmethod
    def get_response_headers(cls) -> dict[str, Any]:
        return {}

    def __init__(self):
        pass

    async def get_allowed_origins(self) -> set[str]:
        return set()

    async def process_request(self, *args, **kwargs) -> None: # type: ignore
        pass

    async def process_response(self, *args, **kwargs) -> None: # type: ignore
        pass


"""Declares :class:`AuthorizationRequest`."""
import dataclasses
import functools
import json
import os
import secrets
import typing

import fastapi
import fastapi.params
import pydantic
from fastapi.params import Query
from pydantic.fields import FieldInfo

from cbra.types import JSONQueryParameter
from .iprincipal import IPrincipal
from .redirecturl import RedirectURL
from .responsetype import ResponseType


class BaseAuthorizationRequest:
    client_id: str
    response_type: ResponseType
    redirect_uri: str | None
    scope: str | None
    state: str | None
    response_mode: str | None
    resource: str | None
    nonce: str | None
    display: str | None
    prompt: str | None
    max_age: int | None
    ui_locales: str | None
    claims_locales: str | None
    id_token_hint: str | None
    login_hint: str | None
    acr_values: str | None
    registration: str | None
    claims: str | None

    @staticmethod
    def get_fields(
        field_factory: typing.Callable[..., FieldInfo | Query],
        private_field: typing.Callable[..., dataclasses.Field[typing.Any] | pydantic.BaseModel],
        exclude: list[str] | None = None
    ) -> dict[str, typing.Any]:
        """Return the authorization request parameters."""
        exclude = exclude or []
        annotations = {
            'client_id': str,
            'response_type': ResponseType,
            'redirect_uri': str | None,
            'scope': str | None,
            'state': str | None,
            'response_mode': str | None,
            'resource': str | None,
            'nonce': str | None,
            'display': str | None,
            'prompt': str | None,
            'max_age': int | None,
            'ui_locales': str | None,
            'claims_locales': str | None,
            'id_token_hint': str | None,
            'login_hint': str | None,
            'acr_values': str | None,
            'registration': str | None,
            'claims': str | None,
            'request_uri': str,
            'request': str
        }
        fields = {
            'client_id': field_factory(
                default=None,
                title="Client identifier",
                description="Identifies the client that is requesting authorization."
            ),
            'response_type': field_factory(
                default=None,
                title="Response type",
                description=(
                    "Informs the authorization server of the desired grant type"
                ),
                example="code",
                enum=[
                    "code",
                    "code id_token",
                    "code id_token token",
                    "code token",
                    "id_token",
                    "id_token token",
                    "none",
                    "token"
                ]
            ),
            'redirect_uri': field_factory(
                default=None,
                title="Redirect URI",
                description=(
                    "The URL to redirect the client to after completing the "
                    "flow. Must be an absolute URI that is served over https.\n\n"
                    "If `redirect_uri` is omitted, the default redirect URI for "
                    "the client specified by `client_id` is used. For clients that "
                    "do not have a redirect URI specified, this produces an error "
                    "state."
                )
            ),
            'scope': field_factory(
                default=None,
                title="Scope",
                description=(
                    "A space-delimited list specifying the requested access scope."
                ),
                example="hello.world"
            ),
            'state': field_factory(
                default=None,
                title="State",
                description=(
                    "An opaque value used by the client to maintain state between "
                    "the request and callback. The authorization server includes "
                    "this value when redirecting the user-agent back to the client."
                ),
                example=bytes.hex(os.urandom(8))
            ),
            'response_mode': field_factory(
                default=None,
                title="Response mode",
                enum=["query", "fragment"],
                description=(
                    "Informs the authorization server of the mechanism to be used "
                    "for returning authorization response parameters."
                ),
                example="query"
            ),
            'resource': field_factory(
                alias='resource',
                default=None,
                title="Resource",
                description=(
                    "Indicates the target service or resource to which access is "
                    "being requested.  Its value MUST be an absolute URI, as "
                    "specified by Section 4.3 of RFC 3986.  The URI MUST NOT "
                    "include a fragment component. It SHOULD NOT include a query "
                    "component, but it is recognized that there are cases that "
                    "make a query component a useful and necessary part of the "
                    "resource parameter, such as when one or more query parameters "
                    "are used to scope requests to an application.  The `resource` "
                    "parameter URI value is an identifier representing the identity "
                    "of the resource, which MAY be a locator that corresponds to a "
                    "network-addressable location where the target resource is "
                    "hosted.  Multiple `resource` parameters MAY be used to indicate "
                    "that the requested token is intended to be used at multiple "
                    "resources."
                ),
                example="https://hello.unimatrixapis.com"
            ),
            'nonce': field_factory(
                default=None,
                title="Nonce",
                description=(
                    "String value used to associate a Client session with an ID "
                    "Token, and to mitigate replay attacks. The value is passed "
                    "through unmodified from the Authentication Request to the "
                    "ID Token. Sufficient entropy MUST be present in the `nonce` "
                    "values used to prevent attackers from guessing values."
                ),
                example=bytes.hex(os.urandom(4))
            ),
            'display': field_factory(
                default=None,
                title="Display",
                description=(
                    "ASCII string value that specifies how the Authorization Server "
                    "displays the authentication and consent user interface pages "
                    "to the End-User. The defined values are:\n\n"
                    "- `page` - The Authorization Server SHOULD display the "
                    "authentication and consent UI consistent with a full User "
                    "Agent page view. If the display parameter is not specified, "
                    "this is the default display mode.\n"
                    "- `popup` - The Authorization Server SHOULD display the "
                    "authentication and consent UI consistent with a popup User "
                    "Agent window. The popup User Agent window should be of an "
                    "appropriate size for a login-focused dialog and should not "
                    "obscure the entire window that it is popping up over.\n"
                    "- `touch` - The Authorization Server SHOULD display the "
                    "authentication and consent UI consistent with a device "
                    "that leverages a touch interface.\n"
                    "- `wap` - The Authorization Server SHOULD display the "
                    "authentication and consent UI consistent with a \"feature "
                    "phone\" type display.\n\n"
                    "If there is no `display` parameter provided, then an attempt "
                    "is made to detect the agent based on the request headers."
                ),
                example="page"
            ),
            'prompt': field_factory(
                default=None,
                title="Prompt",
                description=(
                    "Space delimited, case sensitive list of ASCII string values "
                    "that specifies whether the Authorization Server prompts the "
                    "End-User for reauthentication and consent. The defined values "
                    "are:\n"
                    "\n- `none` - The Authorization Server does not display any "
                    "authentication or consent user interface pages. An error is "
                    "returned if an End-User is not already authenticated or the "
                    "Client does not have pre-configured consent for the requested "
                    "Claims or does not fulfill other conditions for processing "
                    "the request. The error code will typically be `login_required`, "
                    "`interaction_required`, or another code defined in Section "
                    "3.1.2.6 of the OpenID Connect Core 1.0 specification. This "
                    "can be used as a method to check for existing authentication "
                    "and/or consent."
                    "\n- `login` - The Authorization Server prompts the End-User "
                    "for reauthentication. If it cannot reauthenticate the End-User, "
                    "it returns an error, typically `login_required`."
                    "\n- `consent` - The Authorization Server prompts the End-User "
                    "for consent before returning information to the Client. If it "
                    "cannot obtain consent, it returns an error, typically "
                    "`consent_required`."
                    "\n- `select_account` - The Authorization Server prompts "
                    "the End-User to select a user account. This enables an "
                    "End-User who has multiple accounts at the Authorization Server "
                    "to select amongst the multiple accounts that they might have "
                    "current sessions for. If it cannot obtain an account selection "
                    "choice made by the End-User, it return an error, typically "
                    "`account_selection_required`."
                ),
                example="prompt"
            ),
            'max_age': field_factory(
                default=None,
                title="Maximum authentication age",
                description=(
                    "Maximum Authentication Age. Specifies the allowable elapsed time in "
                    "seconds since the last time the End-User was actively authenticated. "
                    "If `max_age` is expired, then the client must re-authenticate."
                )
            ),
            'ui_locales': field_factory(
                default=None,
                title="Preferred locales",
                description=(
                    "End-User's preferred languages and scripts for the user interface, "
                    "represented as a space-separated list of BCP47 language tag values, "
                    "ordered by preference."
                ),
                example="nl-NL be"
            ),
            'claims_locales': field_factory(
                default=None,
                title="Preferred locales for claims",
                description=(
                    "End-User's preferred languages and scripts for returned Claims, "
                    "represented as a space-separated list of BCP47 language tag values, "
                    "ordered by preference."
                ),
                example="nl-NL be"
            ),
            'id_token_hint': field_factory(
                default=None,
                title="ID Token hint",
                description=(
                    "Token previously issued by the Authorization Server being passed as "
                    "a hint about the End-User's current or past authenticated session with "
                    "the Client. If the End-User identified by the ID Token is logged in "
                    "or is logged in by the request, then the Authorization Server returns "
                    "a positive response; otherwise, it return an error, such as "
                    "`login_required`. When possible, an `id_token_hint` should be present "
                    "when `prompt=none` is used and an `invalid_request` error is returned "
                    "if it is not; however, the server attempts to respond successfully "
                    "when possible, even if it is not present. The Authorization Server "
                    "need not be listed as an audience of the ID Token when it is used "
                    "as an `id_token_hint` value."
                )
            ),
            'login_hint': field_factory(
                default=None,
                title="Login hint",
                description=(
                    "Hint to the Authorization Server about the login identifier the End-User "
                    "might use to log in (if necessary). This hint can be used by an RP if it "
                    "first asks the End-User for their e-mail address (or other identifier) "
                    "and then wants to pass that value as a hint to the discovered "
                    "authorization service. It is recommended that the hint value match the "
                    "value used for discovery. This value may also be a phone number in the "
                    "format specified for the `phone_number` Claim."
                ),
                example="hello@unimatrixone.io"
            ),
            'acr_values': field_factory(
                default=None,
                title="ACR values",
                description=(
                    "Requested **Authentication Context Class Reference** values. Space-separated "
                    "string that specifies the `acr` values that the Authorization Server is "
                    "being requested to use for processing this Authentication Request, with the "
                    "values appearing in order of preference. The **Authentication Context Class** "
                    "satisfied by the authentication performed is returned as the `acr` Claim Value. "
                    "The `acr` Claim is requested as a *Voluntary Claim* by this parameter."
                )
            ),
            'registration': field_factory(
                default=None,
                title="Registration",
                description=(
                    "This parameter is used by the Client to provide information about itself to a "
                    "Self-Issued OP that would normally be provided to an OP during Dynamic Client "
                    "Registration. The value is a JSON object containing Client metadata values, as "
                    "defined in Section 2.1 of the OpenID Connect Dynamic Client Registration 1.0 "
                    "specification. The registration parameter must not be used when the OP is not "
                    "a Self-Issued OP."
                )
            ),
            'claims': field_factory(
                default='{}',
                title="Claims",
                description=(
                    "This parameter is used to request that specific Claims be returned. The value is "
                    "a JSON object listing the requested Claims.\n\n"
                    "The claims Authentication Request parameter requests that specific Claims be "
                    "returned from the UserInfo Endpoint and/or in the ID Token. It is represented "
                    "as a JSON object containing lists of Claims being requested from these "
                    "locations. Properties of the Claims being requested MAY also be specified.\n\n"
                    "The claims parameter value is represented in an OAuth 2.0 request as UTF-8 "
                    "encoded JSON (which ends up being form-urlencoded when passed as an OAuth "
                    "parameter). When used in a **Request Object** value, the JSON is used as "
                    "the value of the `claims` member.\n\n"
                    "The top-level members of the Claims request JSON object are:\n\n"
                    "- `userinfo` - (Optional) Requests that the listed individual Claims be "
                    "returned from the UserInfo Endpoint. If present, the listed Claims are "
                    "being requested to be added to any Claims that are being requested using "
                    "scope values. If not present, the Claims being requested from the UserInfo "
                    "Endpoint are only those requested using `scope` values."
                    "\n- `id_token` - (Optional) Requests that the listed individual Claims be "
                    "returned in the ID Token. If present, the listed Claims are being requested "
                    "to be added to the default Claims in the ID Token. If not present, the default "
                    "ID Token Claims are requested.\n\n"
                    "Other members may be present. Any members used that are not understood "
                    "must be ignored."
                )
            ),
            'request': field_factory(
                default=None,
                title="Request",
                description=(
                    "A JSON Web Token (JWT) whose JWT Claims Set holds the "
                    "JSON-encoded OAuth 2.0 authorization request parameters. "
                    "Must not be used in combination with the `request_uri` "
                    "parameter, and all other parameters except `client_id` "
                    "must be absent.\n\n"
                    "Confidential and credentialed clients must first sign "
                    "the claims using their private key, and then encrypt the "
                    "result with the public keys that are provided by the "
                    "authorization server through the `jwks_uri` specified "
                    "in its metadata.\n\n"
                    "A non-standard extension used by this authorization server "
                    "is the ability for public clients to use the `request` "
                    "parameter. In that case, the JWT must be encrypted using "
                    "the method described above, and not signed as a public "
                    "client does not have a key."
                )
            ),
            'request_uri':  field_factory(
                default=None,
                title="Request URI",
                description=(
                    "References a Pushed Authorization Request (PAR) or a remote "
                    "object containing the authorization request.\n\n"
                    "If the authorization request was pushed to this authorization "
                    "server, then the format of the `request_uri` parameter is "
                    "`urn:ietf:params:oauth:request_uri:<reference-value>`. "
                    "Otherwise, it is an URI using the `https` scheme. If the "
                    "`request_uri` parameter is a remote object, then the external "
                    "domain must have been priorly whitelisted by the client."
                )
            )
        }
        for name in exclude:
            if name in fields:
                del fields[name]
            if name in annotations:
                del annotations[name]
        return {
            **fields,
            '__annotations__': annotations
        }

    @classmethod
    def create_model(
        cls,
        location: str,
        exclude: list[str] | None = None
    ) -> type['BaseAuthorizationRequest']:
        """Create a new :class:`BaseAuthorizationRequest` implementation
        configured with the given field type.
        """
        bases: list[typing.Any] = [cls]
        Field = fastapi.Query
        PrivateField = functools.partial(dataclasses.field, init=False)
        if location == 'body':
            Field = pydantic.Field
            PrivateField = pydantic.PrivateAttr
            bases.insert(0, pydantic.BaseModel)
        new_class = type(
            cls.__name__,
            tuple(bases),
            cls.get_fields(
                field_factory=Field,
                private_field=PrivateField,
                exclude=exclude
            )
        )
        if location == 'query':
            new_class = dataclasses.dataclass(new_class)
        return new_class

    @classmethod
    def body(
        cls,
        exclude: list[str] | None = None
    ) -> type['BaseAuthorizationRequest']:
        return cls.create_model('body', exclude=exclude or [])

    @classmethod
    def query(cls) -> type['BaseAuthorizationRequest']:
        return cls.create_model('query')

    @staticmethod
    def space_separated(
        value: str | set[str] | None,
        ordered: bool = False
    ) -> set[str] | list[str]:
        if isinstance(value, (set, list)):
            return value
        if value is None:
            return set()
        values = [x.strip() for x in str.split(value, ' ')]
        if not ordered:
            values = set(values)
        return values

    @property
    def authorization_code(self) -> str | None:
        return self._authorization_code

    def __init__(self, *args: typing.Any, **kwargs: typing.Any) -> None:
        super().__init__(*args, **kwargs)
        if self.needs_authorization_code(): # type: ignore
            self._authorization_code = secrets.token_urlsafe(24)

    def get_acr_values(self) -> set[str]:
        return typing.cast(set[str], self.space_separated(self.acr_values))

    def get_claims(self) -> dict[str, typing.Any]:
        return json.loads(self.claims or "null")

    def get_claims_locales(self) -> list[str]:
        return typing.cast(list[str], self.space_separated(self.claims_locales, True))

    def get_redirect_url(self) -> RedirectURL:
        return RedirectURL.validate(self.redirect_uri)

    def get_resources(self) -> set[str]:
        return typing.cast(set[str], self.space_separated(self.resource))

    def get_scope(self) -> set[str]:
        return typing.cast(set[str], self.space_separated(self.scope))

    def get_principal(self) -> IPrincipal | None:
        """Return the principal specified by the request parameters. For
        requests to the authorization endpoint, this method always returns
        ``None``.
        """
        return None

    def get_ui_locales(self) -> list[str]:
        return typing.cast(list[str], self.space_separated(self.ui_locales, True))

    def is_external(self) -> bool:
        """If the request is a reference, return a boolean indicating if the
        request is made available at an external source.
        """
        return False

    def is_reference(self) -> bool:
        """Return a boolean indicating if the authorization request is a
        reference.
        """
        return False

    class Config:
        check_fields = False
        exclude_defaults = True


"""Declares :class:`AuthorizationRequest`."""
import secrets
import typing
from typing import Any

import pydantic
from ckms.types import JSONWebKey

from ..exceptions import ClientUnsupportedResponseMode
from ..exceptions import Error
from ..exceptions import ForbiddenScope
from ..exceptions import InvalidGrant
from ..exceptions import UnsupportedResponseMode
from ..exceptions import UnsupportedResponseType
from .authorizationcode import AuthorizationCode
from .authorizationcodegrant import AuthorizationCodeGrant
from .authorizationrequestparameters import AuthorizationRequestParameters
from .iclient import IClient
from .isubject import ISubject
from .isubjectrepository import ISubjectRepository
from .iopenidtokenbuilder import IOpenIdTokenBuilder as OIDCHandler
from .redirecturl import RedirectURL
from .responsetype import ResponseType
from .jar import JAR
from .servermetadata import ServerMetadata


class AuthorizationRequest(pydantic.BaseModel):
    """Represents the cleaned parameters of an authorization request."""
    __module__: str = 'cbra.ext.oauth2.types'

    #: Identifies the request. Is used in both Pushed Authorization Requests
    #: (PAR), as well as the `state` parameter (or similar) when the
    #: authorization server redirects the user-agent to an upstream identity
    #: provider. This attribute must never be set manually.
    request_id: str

    #: If the Authorization Code Flow is used, the authorization code that the
    #: client may used to obtain an access token.
    authorization_code: str = pydantic.Field(default='')

    client_id: str
    response_type: ResponseType | None
    redirect_uri: RedirectURL
    scope: set[str]
    state: str | None

    # The Response Mode determines how the Authorization Server returns
    # result parameters from the Authorization Endpoint. Non-default
    # modes are specified using the response_mode request parameter.
    # If response_mode is not present in a request, the default Response
    # Mode mechanism specified by the Response Type is used.
    # (OAuth 2.0 Multiple Response Type Encoding Practices)
    response_mode: str | None
    resource: set[str]
    nonce: str | None
    display: str | None
    prompt: str | None
    max_age: int | None
    ui_locales: list[str]
    claims_locales: list[str]
    id_token_hint: str | None
    login_hint: str | None
    acr_values: set[str]
    registration: str | None
    claims: dict[str, Any] | None

    # Extensions
    access_token: str | None
    id_token: str | None
    encryption_key: JSONWebKey | None

    # Internal
    auth_time: int | None = None
    sub: str | int | None = None
    redirect_uri_provided: bool
    session: dict[str, Any] = {}

    @classmethod
    def fromparams(
        cls,
        params: AuthorizationRequestParameters | JAR
    ) -> 'AuthorizationRequest':
        """Create a new :class:`AuthorizationRequest` from the parameters 
        provided to the :term:`Authorization Endpoint` or :term:`PAR Endpoint`.
        """
        return cls(
            request_id=secrets.token_urlsafe(24),
            client_id=params.client_id,
            response_type=params.response_type,
            redirect_uri=params.get_redirect_url(),
            scope=params.get_scope(),
            state=params.state,
            response_mode=params.response_mode,
            resource=params.get_resources(),
            nonce=params.nonce,
            display=params.display,
            prompt=params.prompt,
            max_age=params.max_age,
            ui_locales=params.get_ui_locales(),
            claims_locales=params.get_claims_locales(),
            id_token_hint=params.id_token_hint,
            login_hint=params.login_hint,
            acr_values=params.get_acr_values(),
            registration=params.registration,
            claims=params.get_claims(),
            access_token=None,
            id_token=None,
            encryption_key=None,
            redirect_uri_provided=bool(params.redirect_uri)
        )

    def __init__(self, **kwargs: typing.Any):
        super().__init__(**kwargs)
        if self.needs_authorization_code():
            self.authorization_code = secrets.token_urlsafe(48)

    #def authorize(self):
    #    """Authorize the request and return a string containing the
    #    redirect URL.
    #    """
    #    params = {
    #        'code': self.authorization_code
    #    }
    #    if self.state is not None:
    #        params['state'] = self.state
    #    assert self.redirect_uri is not None # nosec
    #    return self.redirect_uri.authorize(**params)

    def can_interact(self) -> bool:
        """Return a boolean indicating if the authorization server or its
        delegated can interact with the resource owner.
        """
        return self.prompt != "none"

    def get_authorization_code(self) -> AuthorizationCode:
        """Return the authorization code."""
        if not self.needs_authorization_code():
            raise ValueError("Invalid response type for authorization code.")
        return AuthorizationCode(
            request_id=self.request_id,
            code=self.authorization_code
        )

    async def get_subject(
        self,
        client: IClient,
        subjects: ISubjectRepository
    ) -> ISubject | None:
        """Return the subject instance, if the authorization request provided
        credentials."""
        if not self.is_authenticated():
            return None
        if self.id_token or self.access_token:
            raise NotImplementedError
        return await subjects.get(client, self.sub)

    def get_response_mode(self) -> str:
        """Return the response mode defined by the request, or default to the response
        mode for the given response type.
        """
        return self.response_mode or 'query'

    def is_authenticated(self) -> bool:
        """Return a boolean indicating if the request was authenticated using
        the :attr:`id_token` or :attr:`access_token` parameters.
        """
        return bool(self.id_token or self.access_token or self.sub)

    def is_openid(self) -> bool:
        """Return a boolean indicating if the authorization request is
        an OpenID Connect request.
        """
        return "openid" in self.scope

    def needs_authorization_code(self) -> bool:
        """Return a boolean indicating if the authorization code flow
        is used.
        """
        return self.response_type in {
            ResponseType.code,
            ResponseType.code_id_token,
            ResponseType.code_id_token_token,
            ResponseType.code_token
        }

    def validate_grant(
        self,
        client: IClient,
        subject: ISubject | None,
        grant: AuthorizationCodeGrant
    ) -> None:
        # The "redirect_uri" parameter is required "if the "redirect_uri"
        # parameter was included in the authorization request as described
        # in Section 4.1.1, and their values MUST be identical."
        # (RFC 6749, 4.1.3)
        if self.redirect_uri_provided and not grant.redirect_uri:
            raise InvalidGrant(
                error_description=(
                    "The authorization request specified the \"redirect_uri\" "
                    "parameter, thus the token request must include the "
                    "same parameter with an identical value."
                )
            )
        if str(self.redirect_uri) != grant.redirect_uri:
            raise InvalidGrant(
                error_description=(
                    "The \"redirect_uri\" parameter must be identical to the "
                    "value provided in the authorization request."
                )
            )

    async def validate_request_parameters(
        self,
        *,
        client: IClient,
        subject: ISubject | None,
        openid: OIDCHandler,
        metadata: ServerMetadata | None = None
    ):
        """Validates the authorization request."""
        if self.response_type is None:
            raise Error(
                error="invalid_request",
                error_description="The response_type parameter is required."
            )
        self.redirect_uri = typing.cast(
            RedirectURL,
            client.get_redirect_url(self.redirect_uri)
        )
        if not client.allows_response_type(self.response_type):
            raise UnsupportedResponseType.as_redirect(self.redirect_uri)
        if not client.allows_response_mode(self.response_mode):
            raise ClientUnsupportedResponseMode.as_redirect(self.redirect_uri)
        if not client.allows_scope(self.scope):
            raise ForbiddenScope
        if metadata is not None:
            if not metadata.supports_response_mode(self.response_mode):
                raise UnsupportedResponseMode.as_redirect(self.redirect_uri)

        if self.claims and not openid.validate_claims(self.claims):
            raise Error(
                error="invalid_request",
                error_description="Invalid claims parameter."
            )


"""Declares :class:`IRenderer`."""
import typing


Renderable = typing.Union[
    typing.Dict[typing.Any, typing.Any],
    typing.List[typing.Any]
]


class IRenderer:
    """All renderers should extend this class, setting the `media_type`
    and `format` attributes, and override the `.render()` method.
    """
    __module__: str = 'cbra.types'
    accepted: str
    media_type: str
    format: typing.Optional[str] = None
    charset: typing.Optional[str] = 'utf-8'
    render_style: str = 'text'
    Renderable: typing.Any = Renderable
    structured: bool = False
    response_media_type: typing.Optional[str] = None
    exact: bool = True

    @classmethod
    def configure(cls, **kwargs: typing.Any) -> typing.Type['IRenderer']:
        return type(cls.__name__, (cls,), kwargs)

    def __init__(self, accepted: str):
        self.accepted = accepted

    def render(
        self,
        data: Renderable,
        renderer_context: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> typing.Union[bytes, str]:
        raise NotImplementedError('Renderer class requires .render() to be implemented')

    def has_content(self) -> bool:
        """Return a boolean indicating if the renderer produces content
        at all.
        """
        raise NotImplementedError

    def get_response_encoding(self) -> str:
        assert self.response_media_type is not None # nosec
        return self.response_media_type


"""Declares :class:`BaseClient`."""
import typing

import fastapi

from cbra.ext import ioc
from .exceptions import ClientDoesNotExist
from .exceptions import InvalidClient
from .types import IClient
from .types import IClientRepository


class BaseClient(IClient):
    """The base class for all OAuth 2.0 client implementations."""
    __module__: str = 'cbra.ext.oauth2'

    @classmethod
    def fromdict(cls, params: dict[str, typing.Any]) -> 'IClient':
        return cls(**params)

    @classmethod
    def fromquery(
        cls,
        clients: type,
        required: bool = False
    ) -> typing.Any:
        """Return a callable that resolves the current client from the
        query parameters.
        """
        async def f(
            client_id: str = fastapi.Query(
                default=... if required else None,
                title="Client ID",
                description=(
                    "A client identifier unique to this authorization server."
                )
            ),
            clients: IClientRepository = ioc.instance(clients)
        ) -> typing.Optional[IClient]:
            if client_id is None and required:
                raise InvalidClient
            try:
                return await clients.get(client_id)
            except ClientDoesNotExist:
                if required:
                    raise
                return None

        return fastapi.Depends(f)


# pylint: skip-file
import httpx
import pytest
import pytest_asyncio
from ckms.core import Keychain

from cbra.ext.oauth2.types import ClientAssertionType
from cbra.ext.oauth2.types import IClient
from cbra.ext.oauth2.types import GrantType
from ...assertedclientmixin import AssertedClientMixin
from ...servertest import ServerTest


class TestClientCredentialsFlow(ServerTest, AssertedClientMixin):
    base_url: str = "https://cbra-oauth2.localhost"
    client_id: str = 'test-client-credentials'
    client_key: str = "confidential1"
    redirect_uri: str = "https://localhost:8000/callback"
    assertion_type: ClientAssertionType = ClientAssertionType.jwt_bearer

    @pytest_asyncio.fixture # type: ignore
    async def rejected_keychain(self) -> Keychain:
        keychain = Keychain()
        keychain.configure({
            'sig': {
                'provider': "local",
                'kty': "OKP",
                'path': "pki/confidential2.key",
                'alg': "EdDSA",
                'crv': "Ed448",
                'use': 'sig'
            }
        })
        await keychain
        return keychain

    @pytest.mark.asyncio
    async def test_assertion_obtain_token(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        dto = response.json()
        assert response.status_code == 200, response.text
        self.assert_rfc9068(dto['access_token'], 'iss', self.base_url)
        self.assert_rfc9068(dto['access_token'], 'sub', self.client_id)
        self.assert_rfc9068(dto['access_token'], 'client_id', self.client_id)
        self.assert_rfc9068(dto['access_token'], 'scope', "system")


    # If the request does not include a "resource" parameter, the authorization
    # server MUST use a default resource indicator in the "aud" claim (RFC 9068,
    # Section 3).
    @pytest.mark.asyncio
    async def test_assertion_sets_default_resource(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        dto = response.json()
        assert response.status_code == 200, response.text
        self.assert_rfc9068(dto['access_token'], 'aud', {self.base_url})

    @pytest.mark.asyncio
    async def test_assertion_sets_allowed_resource(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        resource = "https://resource1.unimatrixapis.localhost"
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system",
                'resource': [resource]
            }
        )
        dto = response.json()
        assert response.status_code == 200, response.text
        self.assert_rfc9068(dto['access_token'], 'aud', {resource})

    @pytest.mark.asyncio
    async def test_assertion_sets_allowed_resources(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient,
        current_client: IClient
    ):
        current_client.allow_multiple_audiences = True
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        resources = [
            "https://resource1.unimatrixapis.localhost",
            "https://resource2.unimatrixapis.localhost"
        ]
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system",
                'resource': resources
            }
        )
        current_client.allow_multiple_audiences = False
        dto = response.json()
        assert response.status_code == 200, response.text
        self.assert_rfc9068(dto['access_token'], 'aud', set(resources))

    @pytest.mark.asyncio
    async def test_assertion_forbids_multiple_resources(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        resources = [
            "https://resource1.unimatrixapis.localhost",
            "https://resource2.unimatrixapis.localhost"
        ]
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system",
                'resource': resources
            }
        )
        assert response.status_code == 400, response.text
        self.assert_error(response, 'invalid_grant')

    @pytest.mark.asyncio
    async def test_assertion_forbided_non_whitelisted_resources(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        resource = "https://resource3.unimatrixapis.localhost"
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system",
                'resource': resource
            }
        )
        assert response.status_code == 400, response.text
        self.assert_error(response, 'invalid_grant')

    @pytest.mark.asyncio
    async def test_assertion_can_not_use_multiple_times(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        assert response.status_code == 200, response.text
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        assert response.status_code == 400, response.text
        self.assert_error(response, 'invalid_client')

    @pytest.mark.asyncio
    async def test_assertion_audience_must_be_token_endpoint(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.base_url,
            sub=self.client_id
        )
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        assert response.status_code == 400, response.text
        self.assert_error(response, "invalid_grant")

    @pytest.mark.asyncio
    async def test_assertion_signature_must_be_from_client(
        self,
        rejected_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=rejected_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub=self.client_id
        )
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        assert response.status_code == 400, response.text
        self.assert_error(response, "invalid_grant")

    # If the client JWT is not valid, the authorization server constructs
    # an error response as defined in OAuth 2.0 [RFC6749].  The value of
    # the "error" parameter MUST be the "invalid_client" error code.
    # The authorization server MAY include additional information regarding
    # the reasons the JWT was considered invalid using the "error_description"
    # or "error_uri" parameters.
    @pytest.mark.asyncio
    async def test_assertion_sub_must_match_client(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss=self.client_id,
            aud=self.token_url,
            sub='adifferentclient'
        )
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        assert response.status_code == 400, response.text
        self.assert_error(response, "invalid_client")

    @pytest.mark.asyncio
    async def test_assertion_iss_must_match_client(
        self,
        client_keychain: Keychain,
        client: httpx.AsyncClient
    ):
        assertion = await self.get_client_assertion(
            keychain=client_keychain,
            iss='adifferentclient',
            aud=self.token_url,
            sub=self.client_id
        )
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'client_assertion_type': self.assertion_type,
                'client_assertion': assertion,
                'scope': "system"
            }
        )
        assert response.status_code == 400, response.text
        self.assert_error(response, "invalid_client")

    # The client MUST authenticate with the authorization server (RFC 6749, 4.4.2)
    @pytest.mark.asyncio
    async def test_client_must_authenticate(
        self,
        client: httpx.AsyncClient
    ):
        response = await client.post( # type: ignore
            url=self.token_url,
            json={
                'grant_type': GrantType.client_credentials,
                'client_id': self.client_id,
                'scope': "system"
            }
        )
        self.assert_error(response, "invalid_client")


# pylint: skip-file
# type: ignore
from cbra import Endpoint


def test_create_sets_attribute():
    cls = Endpoint.new(foo=1)
    assert hasattr(cls, 'foo')
    assert cls.foo == 1


def test_create_overrides_attribute():
    cls = Endpoint.new(principal=1)
    assert cls.principal == 1



# pylint: skip-file
# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import AsyncGenerator

import pytest_asyncio
from httpx import AsyncClient

from cbra import Application
from examples.woocommerce import get_asgi_application


@pytest_asyncio.fixture # type: ignore
async def app() -> Application:
    app = get_asgi_application()
    await app.on_startup()
    return app


@pytest_asyncio.fixture # type: ignore
async def client(
    app: Application,
    base_url: str
) -> AsyncGenerator[AsyncClient, None]:
    params = {
        'app': app,
        'base_url': base_url
    }
    async with AsyncClient(**params) as client:
        yield client


# pylint: skip-file
import typing
import urllib.parse

import httpx

from ...assertedclientmixin import AssertedClientMixin
from ...jarmixin import JARMixin
from ...servertest import ServerTest
from .iauthorizationcodeflow import IAuthorizationCodeFlow


class TestJARFlow(ServerTest, JARMixin, IAuthorizationCodeFlow, AssertedClientMixin):
    client_id: str = 'test-jar'
    client_key: str = 'confidential1'
    cors_allowed_methods: set[str] = {"POST"}
    cors_url: str = ServerTest.jar_url
    redirects: bool = False
    response_type: str = "code"

    async def get_token_client_params(self) -> dict[str, typing.Any]:
        return {
            'client_id': self.client_id,
            'client_assertion_type': self.assertion_type,
            'client_assertion': await self.get_client_assertion(
                keychain=await self.get_client_keychain(self.client_key),
                sub=self.client_id
            )
        }

    async def create_authorize_response(
        self,
        client: httpx.AsyncClient,
        params: dict[str, str],
        status_code: int = 303,
        error: str | None = None,
        headers: dict[str, str] | None = None
    ) -> tuple[urllib.parse.ParseResult, dict[str, str], dict[str, str]]:
        try:
            request_uri, _ = await self.create_jar_response(
                client=client,
                params=params,
                codec=await self.get_client_codec(
                    keychain=await self.get_client_keychain(self.client_key)
                ),
                headers=headers
            )
        except httpx.HTTPStatusError as exception:
            if error is None:
                raise
            dto = exception.response.json()
            assert exception.response.status_code == status_code
            assert dto.get('error') == error
            return (urllib.parse.urlparse(''), dict(exception.response.headers), {})

        params = {
            'client_id': self.client_id,
            'request_uri': request_uri
        }
        return await super().create_authorize_response(client, params, status_code, error)


class TestJARFlowCORS(TestJARFlow):
    origin = "https://allowed-cors-domain.example"


# pylint: skip-file
import asyncio
import functools

import fastapi
import pytest
import pytest_asyncio
from fastapi.concurrency import AsyncExitStack
from httpx import AsyncClient
from ioc.provider import Provider as DependencyContainer
from unimatrix.ext.kms import Keychain

import cbra


TEST_HMAC_KEY: str = '8a983a3d5282c751feadbdd1ad23398f882f65046df4b0ac5a96c93a50669ff7'

TEST_DEPENDENCY = "Foo"


@pytest.fixture(scope='session')
def event_loop():
    return asyncio.get_event_loop()


@pytest_asyncio.fixture(scope='session')
async def keychain():
    keychain = Keychain()
    await keychain.register('hmac', [f"literal+hmac:{TEST_HMAC_KEY}?keyid=hmac"])
    await keychain.register('rsa', ['file+pem:pki/rsa.key'])
    await keychain.register('p256', ['file+pem:pki/p256.key'])
    await keychain.register('p384', ['file+pem:pki/p384.key'])
    await keychain.register('p521', ['file+pem:pki/p521.key'])

    return keychain


@pytest_asyncio.fixture
async def run_dependant(app):
    runner = cbra.DependantRunner()
    async with AsyncExitStack() as stack:
        request = fastapi.Request(
            scope={
                'app': app,
                'type': "http",
                'query_string': '',
                'headers': [
                    (b'host', b'foo.bar.baz')
                ],
                'fastapi_astack': stack
            }
        )
        yield functools.partial(runner.run_dependant, request)


@pytest_asyncio.fixture
async def client(app):
    async with AsyncClient(app=app, base_url="http://test") as client:
        yield client


@pytest.fixture
def container():
    return DependencyContainer()


@pytest.fixture
def test_dependency():
    return TEST_DEPENDENCY



"""Declares :class:`NamedQueryMetaclass`."""
import typing

from .namedqueryargs import NamedQueryArgs


class NamedQueryMetaclass(type):
    reserved_attrs: set = {
        "build",
        "resolve",
        "run",
    }

    def __new__(cls, name, bases, attrs):
        new = super().__new__
        if attrs.get('__abstract__', False):
            return new(cls, name, bases, attrs)
        attrs['__abstract__'] = False
        hints = attrs.get('__annotations__') or {}
        fields = {}
        for k in dict.keys(hints): # pragma: no cover
            if str.startswith(k, '_'):
                continue
            if k in cls.reserved_attrs:
                raise ValueError(f"Attribute `{k}` is a reserved name.")
            if k in attrs:
                fields[k] = attrs.pop(k)
        attrs['parameterless'] = bool(fields)
        attrs['model'] = cls.create_model(
            cls,
            name,
            cls.get_model_bases(cls, bases),
            hints,
            fields
        )

        return new(cls, name, bases, attrs)

    def get_model_bases(cls, bases: typing.List[type]) -> typing.Tuple[type]:
        """Return the bases that are used to construct the query model."""
        parents = [
            b for b in bases
            if isinstance(b, NamedQueryMetaclass)
            and not b.__abstract__
        ]
        return (NamedQueryArgs,) if not parents else (parents[0].model,)

    def create_model(cls,
        name: str,
        bases: typing.List[type],
        hints: dict,
        fields: dict
    ) -> NamedQueryArgs:
        """Construct the model used to validate the query parameters."""
        return type(name, cls.get_model_bases(cls, bases), {
            '__annotations__': hints,
            **fields
        })



"""Declares :class:`AuthorizationEndpointTest`."""
from .base import BaseOpenAuthorizationTest


class AuthorizationEndpointTest(BaseOpenAuthorizationTest):
    __module__: str = 'cbra.ext.oauth2.tests'
    pass



# pylint: skip-file
# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest
from httpx import AsyncClient


@pytest.mark.asyncio
async def test_ping(
    client: AsyncClient
) -> None:
    response = await client.post( # type: ignore
        url='/',
        headers={
            'Content-Type': 'application/x-www-form-urlencoded'
        },
        content=b'webhook_id=1'
    )
    assert response.status_code == 200, response.text


@pytest.mark.asyncio
async def test_ping_invalid_schema(
    client: AsyncClient
) -> None:
    response = await client.post( # type: ignore
        url='/',
        headers={
            'Content-Type': 'application/x-www-form-urlencoded'
        },
        content=b'foo_id=1'
    )
    assert response.status_code == 422


@pytest.mark.asyncio
async def test_ping_invalid_value(
    client: AsyncClient
) -> None:
    response = await client.post( # type: ignore
        url='/',
        headers={
            'Content-Type': 'application/x-www-form-urlencoded'
        },
        content=b'webhook_id=a'
    )
    assert response.status_code == 422


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# pylint: skip-file
# type: ignore
from typing import Any

from .basemessagetest import BaseMessageTest


class TestProductRestored(BaseMessageTest):
    event_type: str = 'restored'
    resource_type: str = 'product'

    def get_request_params(self) -> dict[str, Any]:
        return {
            'headers': {
                'content-type': "application/json",
                'x-wc-webhook-delivery-id': "89d25cfbe5fb300cab823bb41ee020d9",
                'x-wc-webhook-event': "restored",
                'x-wc-webhook-id': "60",
                'x-wc-webhook-resource': "product",
                'x-wc-webhook-signature': "olvOF/mmu05Z4NmHGoJRg334+aKJRO+irbp/5OB4Ijc=",
                'x-wc-webhook-source': "https://test-wordpress.molano.nl/",
                'x-wc-webhook-topic': "product.restored",
            },
            'content': (
                b'{"id":74,"name":"product.created","slug":"product-cr'
                b'eated","permalink":"https:\/\/test-wordpress.molano.'
                b'nl\/product\/product-created\/","date_created":"2022'
                b'-06-03T17:22:45","date_created_gmt":"2022-06-03T17:2'
                b'2:45","date_modified":"2022-06-03T18:22:12","date_mo'
                b'dified_gmt":"2022-06-03T18:22:12","type":"simple","s'
                b'tatus":"publish","featured":false,"catalog_visibilit'
                b'y":"visible","description":"","short_description":""'
                b',"sku":"","price":"","regular_price":"","sale_price"'
                b':"","date_on_sale_from":null,"date_on_sale_from_gmt"'
                b':null,"date_on_sale_to":null,"date_on_sale_to_gmt":n'
                b'ull,"on_sale":false,"purchasable":false,"total_sales'
                b'":0,"virtual":false,"downloadable":false,"downloads"'
                b':[],"download_limit":-1,"download_expiry":-1,"extern'
                b'al_url":"","button_text":"","tax_status":"taxable","'
                b'tax_class":"","manage_stock":false,"stock_quantity":'
                b'null,"backorders":"no","backorders_allowed":false,"b'
                b'ackordered":false,"low_stock_amount":null,"sold_indi'
                b'vidually":false,"weight":"","dimensions":{"length":"'
                b'","width":"","height":""},"shipping_required":true,"'
                b'shipping_taxable":true,"shipping_class":"","shipping'
                b'_class_id":0,"reviews_allowed":true,"average_rating"'
                b':"0.00","rating_count":0,"upsell_ids":[],"cross_sell'
                b'_ids":[],"parent_id":0,"purchase_note":"","categorie'
                b's":[{"id":15,"name":"Uncategorized","slug":"uncatego'
                b'rized"}],"tags":[],"images":[],"attributes":[],"defa'
                b'ult_attributes":[],"variations":[],"grouped_products'
                b'":[],"menu_order":0,"price_html":"","related_ids":[4'
                b'6,44,13,39,14],"meta_data":[],"stock_status":"instoc'
                b'k","_links":{"self":[{"href":"https:\/\/test-wordpre'
                b'ss.molano.nl\/wp-json\/wc\/v3\/products\/74"}],"coll'
                b'ection":[{"href":"https:\/\/test-wordpress.molano.nl'
                b'\/wp-json\/wc\/v3\/products"}]}}'
            )
        }


"""Declares :class:`OptionsResponse`."""
from typing import cast
from typing import Any

from starlette.background import BackgroundTask
from starlette.responses import Response


class OptionsResponse(Response):

    def __init__(
        self,
        content: Any = None,
        status_code: int = 200,
        headers: dict[str, str] | None = None,
        media_type: str | None = None,
        background: BackgroundTask | None = None
    ) -> None:
        super().__init__( # type: ignore
            content=b"",
            status_code=status_code,
            headers=headers or {},
            background=cast(BackgroundTask, background)
        )


"""Declares :class:`ICanonicalException`."""
import fastapi

from .icontentnegotiation import IContentNegotiation
from .irenderer import IRenderer
from .irequesthandler import IRequestHandler


class ICanonicalException:

    async def handle(
        self,
        request: fastapi.Request,
        handler: IRequestHandler,
        negotiation: IContentNegotiation,
        renderer: IRenderer
    ) -> fastapi.Response:
        raise NotImplementedError


"""Declares :class:`NullContentNegotiation`."""
from typing import Any

from cbra.parsers import NullParser
from cbra.renderers import NullRenderer
from cbra.types import IParser
from cbra.types import IRenderer
from .default import DefaultContentNegotiation


class NullContentNegotiation(DefaultContentNegotiation):
    """Completely disables all forms of content negotiation."""

    @classmethod
    def get_response_headers(cls) -> dict[str, Any]:
        return {}

    @classmethod
    def has_response_body(cls) -> bool:
        return False

    @classmethod
    def with_body(cls) -> 'NullContentNegotiation': # type: ignore
        return cls()

    def __init__(self):
        pass

    def select_parser(self, *args, **kwargs) -> IParser: # type: ignore
        return NullParser()

    def select_renderer(self, *args, **kwargs) -> IRenderer: # type: ignore
        return NullRenderer("*/*")


"""Declares :class:`YAMLParser`."""
import typing

import fastapi
import yaml

from cbra.types import IParser
from ..exceptions import ParseError


class YAMLParser(IParser):
    """Parses YAML-serialized data."""
    __module__: str = 'cbra.parsers'
    media_type: str = 'application/yaml'
    charset: str    = "utf-8"

    async def parse(self,
        request: fastapi.Request,
        media_type: typing.Optional[str] = None,
        parser_context: typing.Optional[typing.Dict[str, typing.Any]] = None
    ) -> typing.Any:
        """Parses the incoming bytestream as YAML and returns the resulting
        data.
        """
        parser_context = parser_context or {}
        encoding = parser_context.get('encoding', self.charset)
        try:
            content = await request.body()
            if content is None:
                return None
            assert content is not None # nosec
            return yaml.safe_load(bytes.decode(content, encoding)) # type: ignore
        except ValueError:
            raise ParseError
